/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.beacon;

import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.exception.ExceptionSafeErrorHandler;
import com.aquenos.epics.jackie.common.exception.JavaUtilLoggingErrorHandler;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConstants;
import com.aquenos.epics.jackie.common.util.EnvironmentUtil;

/**
 * Configuration for a {@link BeaconDetector}.
 * 
 * @author Sebastian Marsching
 */
public class BeaconDetectorConfiguration {

    private ErrorHandler errorHandler;
    private int defaultServerPort;
    private int repeaterPort;

    /**
     * Creates a configuration for a beacon detector using the
     * <code>EPICS_CA_REPEATER_PORT</code> and <code>EPICS_CA_SERVER_PORT</code>
     * environment variables. These variables have exactly the same meaning and
     * default values as specified by the Channel Access Reference Manual.
     */
    public BeaconDetectorConfiguration() {
        this(null, null, null);
    }

    /**
     * Creates a beacon detector configuration using the specified options. Each
     * option that is <code>null</code> is initialized from its corresponding
     * environment variable.
     * 
     * @param errorHandler
     *            error handler which is used to log errors that are encountered
     *            by the beacon detector or <code>null</code> if a
     *            {@link JavaUtilLoggingErrorHandler} shall be used.
     * @param defaultServerPort
     *            default port to assume for servers if no port is specified
     *            explicitly or <code>null</code> to use
     *            <code>EPICS_CA_SERVER_PORT</code>.
     * @param repeaterPort
     *            port on which the repeater is listening or <code>null</code>
     *            to use <code>EPICS_CA_REPEATER_PORT</code>.
     * @throws IllegalArgumentException
     *             if <code>repeaterPort</code> is specified and zero or an
     *             invalid port number.
     */
    public BeaconDetectorConfiguration(ErrorHandler errorHandler,
            Integer defaultServerPort, Integer repeaterPort) {
        if (errorHandler != null) {
            this.errorHandler = new ExceptionSafeErrorHandler(errorHandler);
        } else {
            this.errorHandler = new ExceptionSafeErrorHandler(
                    new JavaUtilLoggingErrorHandler());
        }
        if (defaultServerPort != null) {
            this.defaultServerPort = defaultServerPort;
            if (defaultServerPort <= 0 || defaultServerPort > 65535) {
                throw new IllegalArgumentException(
                        "Invalid default server port: "
                                + this.defaultServerPort);
            }
        } else {
            initializeDefaultServerPortFromEnvironment();
        }
        if (repeaterPort != null) {
            this.repeaterPort = repeaterPort;
            if (repeaterPort <= 0 || repeaterPort > 65535) {
                throw new IllegalArgumentException("Invalid repeater port: "
                        + this.repeaterPort);
            }
        } else {
            initializeRepeaterPortFromEnvironment();
        }
    }

    private void initializeDefaultServerPortFromEnvironment() {
        defaultServerPort = EnvironmentUtil.portNumberFromEnvironment(
                "EPICS_CA_SERVER_PORT",
                ChannelAccessConstants.DEFAULT_SERVER_PORT, false);
    }

    private void initializeRepeaterPortFromEnvironment() {
        repeaterPort = EnvironmentUtil.portNumberFromEnvironment(
                "EPICS_CA_REPEATER_PORT",
                ChannelAccessConstants.DEFAULT_REPEATER_PORT, false);
    }

    /**
     * Returns the error handler that should be used by the beacon detector.
     * This configuration object ensures that the error handler returned will
     * never throw an exception by wrapping it in a
     * {@link ExceptionSafeErrorHandler}.
     * 
     * @return error handler that is exception safe.
     */
    public ErrorHandler getErrorHandler() {
        return errorHandler;
    }

    /**
     * Returns the TCP and UDP port that should be used when connecting to a
     * server for name resolution and no port has been specified for this server
     * explicitly.
     * 
     * @return default server port for name resolution.
     */
    public int getDefaultServerPort() {
        return defaultServerPort;
    }

    /**
     * Returns the UDP port number that should be used when connecting to the
     * Channel Access repeater.
     * 
     * @return UDP port number for the Channel Access repeater.
     */
    public int getRepeaterPort() {
        return repeaterPort;
    }

}
