/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.internal;

import java.util.concurrent.ExecutionException;

import com.aquenos.epics.jackie.client.ClientThreadingStrategy;
import com.aquenos.epics.jackie.client.DefaultChannelAccessClient;
import com.aquenos.epics.jackie.common.exception.ChannelAccessException;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.util.FutureCompletionListener;
import com.aquenos.epics.jackie.common.util.IntegerIdPool;
import com.aquenos.epics.jackie.common.util.ListenableFuture;
import com.aquenos.epics.jackie.common.util.ListenerLockPolicy;
import com.aquenos.epics.jackie.common.util.SimpleListenableFuture;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Specialization of a {@link ListenableFuture} that is used for read
 * operations. This class is only intended for use by the classes inside this
 * package and by the {@link DefaultChannelAccessClient}.
 * </p>
 * 
 * <p>
 * In addition to the normal features of a {@link ListenableFuture}, this
 * implementation stores some meta-data associated with the read request and
 * supports a way cancel the operation early (typically because the connection
 * to the server has been lost).
 * </p>
 * 
 * <p>
 * It catches all exceptions thrown by its listeners and logs them using the
 * supplied {@link ErrorHandler}, so that they do not bubble up into the calling
 * code.
 * </p>
 * 
 * @author Sebastian Marsching
 *
 * @param <V>
 *            The exact type of the Channel Access values that are provided as a
 *            result of the read request.
 */
public class ReadOperation<V> extends SimpleListenableFuture<V> {

    private ErrorHandler errorHandler;
    private IntegerIdPool.Id operationId;
    private ChannelAccessValueType requestedDataType;
    private int requestedCount;
    private ClientThreadingStrategy threadingStrategy;

    /**
     * Creates a read-operation future.
     * 
     * @param errorHandler
     *            error handler to be used when a listener registered to this
     *            future throws an exception.
     * @param listenerLockPolicy
     *            listener lock-policy to be used by this future.
     * @param operationId
     *            client-assigned ID uniquely identifying this operation.
     * @param requestedDataType
     *            data-type that shall be used in the response from the server.
     * @param requestedCount
     *            number of elements that shall be sent in the response from the
     *            server.
     * @param threadingStrategy
     *            threading strategy used by the client. This operation uses the
     *            threading strategy when notifying completion listeners.
     */
    public ReadOperation(ErrorHandler errorHandler,
            ListenerLockPolicy listenerLockPolicy, IntegerIdPool.Id operationId,
            ChannelAccessValueType requestedDataType, int requestedCount,
            ClientThreadingStrategy threadingStrategy) {
        super(listenerLockPolicy);
        if (operationId == null || requestedDataType == null) {
            throw new NullPointerException();
        }
        this.errorHandler = errorHandler;
        this.operationId = operationId;
        this.requestedDataType = requestedDataType;
        this.requestedCount = requestedCount;
        this.threadingStrategy = threadingStrategy;
    }

    /**
     * <p>
     * Cancels this operation if it has not finished yet. The {@link #get()}
     * method will throw an {@link ExecutionException} wrapping a
     * {@link ChannelAccessException} with a status of
     * {@link ChannelAccessStatus.StatusMessage#ECA_DISCONN ECA_DISCONN}.
     * </p>
     * 
     * <p>
     * This does not release the operation ID because the operation ID may not
     * be reused until there is a reply from the server (or the server
     * connection has been closed).
     * </p>
     */
    public void operationCancelled() {
        if (isDone()) {
            return;
        }
        try {
            setExecutionException(new ChannelAccessException(
                    ChannelAccessStatus.StatusMessage.ECA_DISCONN.toStatus()));
        } catch (IllegalStateException e) {
            // The operation finished before we could indicate that the channel
            // was disconnected. This is not an error.
        }
    }

    /**
     * Returns the operation ID associated with this request. This is the ID
     * that uniquely identifies the operation within the server connection.
     * 
     * @return operation ID.
     */
    public IntegerIdPool.Id getOperationId() {
        return operationId;
    }

    /**
     * Returns the data type that has been requested from the server. The server
     * is expected to reply with this data type.
     * 
     * @return data type requested from the server.
     */
    public ChannelAccessValueType getRequestedDataType() {
        return requestedDataType;
    }

    /**
     * Returns the element count that has been requested from the server. The
     * server is expected to reply with this number of elements.
     * 
     * @return number of elements requested from the server.
     */
    public int getRequestedCount() {
        return requestedCount;
    }

    @Override
    protected void notifyCompletionListener(
            FutureCompletionListener<? super V> listener) {
        try {
            threadingStrategy.notifyFutureCompletionListener(this, listener);
        } catch (Throwable e) {
            errorHandler.handleError(this.getClass(), e,
                    "The completion listener of read-operation future threw an exception.");
        }
    }

}
