/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.resolver.internal;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * <p>
 * Timer that sends search requests at a certain interval. The algorithm used by
 * this class is closely modeled after the algorithm used by the C library so
 * that a similar network performance can be expected.
 * </p>
 * 
 * <p>
 * For each search interval, there is a corresponding object of this class. Each
 * objects keeps a list of channel search requests, that it processes. It sends
 * out messages for this search requests at its defined interval. If a search
 * request is not successful (there is no response until the next run of the
 * timer), it is passed on to the timer with the next larger interval, until it
 * has reached the timer with the largest interval.
 * </p>
 * 
 * <p>
 * There is some logic through which search requests can fall back into a timer
 * with a smaller interval. This logic is triggered when a search request for a
 * timer was successful or when a beacon anomaly is detected.
 * </p>
 * 
 * <p>
 * This class is only intended for use by {@link DefaultChannelNameResolverImpl}
 * . It is closely tied to that class and has only been separated into a class
 * of its own to keep that class smaller.
 * </p>
 * 
 * <p>
 * This class is not thread safe, but the instantiating
 * {@link DefaultChannelNameResolverImpl} ensures that it is only used in a safe
 * way.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class SearchTimer {

    /**
     * Possible return codes for
     * {@link SearchTimer#addSearchMessage(ChannelSearchRequest)}.
     * 
     * @author Sebastian Marsching
     */
    private static enum AddSearchResult {
        /**
         * Message was added successfully.
         */
        SUCCESS,

        /**
         * Message was not added because the space in the packet is insufficient
         * (or the message is too large).
         */
        PACKET_FULL,

        /**
         * Message was not added because an exception occurred. This typically
         * indicates a problem with the search request.
         */
        EXCEPTION
    }

    /**
     * Lower limit for {@link #framesPerTry}.
     */
    private static final double MIN_FRAMES_PER_TRY = 1.0;

    /**
     * Upper limit for {@link #framesPerTry}.
     */
    private static final double MAX_FRAMES_PER_TRY = 64.0;

    /**
     * Maximum number of frames that is sent each time the {@link #process()}
     * method is called.
     */
    private double framesPerTry = MIN_FRAMES_PER_TRY;

    /**
     * Congestion threshold for {@link #framesPerTry} this effectively limits
     * how large {@link #framesPerTry} can grow.
     */
    private double framesPerTryCongestionThreshold = Double.MAX_VALUE;

    /**
     * UDP sequence number at the start of the last run (smallest sequence
     * number sent).
     */
    private int sequenceNumberAtBegin = 0;

    /**
     * UDP sequence number at the end of the last run (largest sequence number
     * sent).
     */
    private int sequenceNumberAtEnd = 0;

    /**
     * Search requests that are waiting to be sent.
     */
    private LinkedHashMap<Integer, ChannelSearchRequest> requestPending = new LinkedHashMap<Integer, ChannelSearchRequest>();

    /**
     * Search requests which have been sent, but for which no response has been
     * received yet.
     */
    private LinkedHashMap<Integer, ChannelSearchRequest> responsePending = new LinkedHashMap<Integer, ChannelSearchRequest>();

    /**
     * System time at the start of the last run.
     */
    private long timeAtLastSend = System.currentTimeMillis();

    /**
     * Have we sent (at least) one search request during the last run?
     */
    private boolean searchRequestSent = false;

    /**
     * Has a response been received since the last run?
     */
    private boolean searchResponseReceived = false;

    /**
     * Is this search timer above the beacon-anomaly timer (has a greater
     * interval). If <code>true</code>, channels can be put back to the
     * beacon-anomaly timer under certain circumstances.
     */
    private boolean boostPossible;

    /**
     * Index of this timer. The timer with the smallest interval has the index
     * zero.
     */
    private int index;

    /**
     * Factor between this timer's search interval and the search interval used
     * by the timer with the smallest interval. This factor is {@link #index} to
     * the power of two.
     */
    private long intervalFactor;

    /**
     * Round-trip-time estimator used for calculating the interval between two
     * runs. The interval between two runs is {@link #intervalFactor} *
     * {@link RoundTripTimeEstimator#getEstimate()}.
     */
    private RoundTripTimeEstimator rttEstimator;

    /**
     * Channel name resolver this search timer belongs to.
     */
    private DefaultChannelNameResolverImpl resolver;

    /**
     * Creates a search timer for a certain search interval.
     * 
     * @param index
     *            index of this timer. For each index, there is exactly one
     *            timer. The index also specifies the search interval.
     * @param boostPossible
     *            <code>true</code> if this timer has a greater index (and thus
     *            search interval) than the beacon-anomaly timer.
     * @param resolver
     *            the channel name resolver that is creating this timer.
     * @param rttEstimator
     *            the round-trip-time estimator that should be used by this
     *            timer when calculating its search interval.
     */
    public SearchTimer(int index, boolean boostPossible,
            DefaultChannelNameResolverImpl resolver,
            RoundTripTimeEstimator rttEstimator) {
        this.index = index;
        this.intervalFactor = 1 << index;
        this.boostPossible = boostPossible;
        this.resolver = resolver;
        this.rttEstimator = rttEstimator;
    }

    /**
     * Adds a search request to this timer. The search request is added to the
     * list of requests that shall be sent when this timer's {@link #process()}
     * method is called.
     * 
     * @param request
     *            search request which is added to this timer.
     */
    public void add(ChannelSearchRequest request) {
        this.requestPending.put(request.getCID(), request);
    }

    /**
     * Removes the specified search request from this timer. The request is
     * always removed, whether it is still queued to be sent or has already been
     * sent and is just waiting for a response.
     * 
     * @param request
     *            request that shall be removed from this timer.
     */
    public void remove(ChannelSearchRequest request) {
        requestPending.remove(request.getCID());
        responsePending.remove(request.getCID());
    }

    /**
     * Removes all search requests from this timer, returning them. The requests
     * are always removed, whether they are still queued to be sent or have
     * already been sent and are just waiting for a response.
     * 
     * @return list of search requests that were formerly registered with this
     *         timer.
     */
    public List<ChannelSearchRequest> removeAll() {
        LinkedList<ChannelSearchRequest> all = new LinkedList<ChannelSearchRequest>();
        all.addAll(responsePending.values());
        responsePending.clear();
        all.addAll(requestPending.values());
        requestPending.clear();
        return all;
    }

    /**
     * Processes pending search requests, sending search messages for them. This
     * method is called periodically by {@link DefaultChannelNameResolverImpl}.
     * The interval is defined by this timer's search interval. Not all search
     * requests might be processed because the timer has a limit on how many UDP
     * packets it will send per run.
     * 
     * @return next time this method should be called in the form of a Java
     *         time-stamp.
     */
    public long process() {
        // Before clearing the list of pending responses, we have to remember
        // whether the list is empty (all requests received a response).
        boolean responseForAllRequests = responsePending.isEmpty();
        // Move all search requests that have not received a response yet to
        // the next larger search period. If this is already the largest search
        // period, this will simply add them back to the requestPending list.
        resolver.moveSearchRequests(index + 1, responsePending.values());
        responsePending.clear();
        // Remember the time when we started to send search requests so that we
        // can update the round-trip-time estimate when we receive replies.
        timeAtLastSend = System.currentTimeMillis();
        // If we received any responses at all, we move all requests that have
        // not been send to the search period that is also used when a beacon
        // anomaly is detected.
        if (searchResponseReceived && boostPossible) {
            resolver.moveSearchRequestsToBeaconAnomalyPeriod(requestPending
                    .values());
            requestPending.clear();
        }
        // If the last time no search requests have been sent, we do not update
        // the frames-per-try settings.
        if (searchRequestSent) {
            if (responseForAllRequests) {
                // If all search requests were successful, we double the number
                // of frames send per interval until we hit the congestion
                // threshold.
                if (framesPerTry < MAX_FRAMES_PER_TRY) {
                    if (framesPerTry < framesPerTryCongestionThreshold) {
                        double doubleFramesPerTry = 2.0 * framesPerTry;
                        if (doubleFramesPerTry > framesPerTryCongestionThreshold) {
                            framesPerTry = framesPerTryCongestionThreshold;
                        } else {
                            framesPerTry = doubleFramesPerTry;
                        }
                    } else {
                        framesPerTry += 1.0 / framesPerTry;
                    }
                }
            } else {
                // If some search requests were not answered, we reduce the
                // congestion threshold and continue with one frame per
                // interval. This should ensure that we do not flood a server's
                // receive buffer.
                framesPerTryCongestionThreshold = framesPerTry / 2.0;
                framesPerTry = 1.0;
            }
        }
        // We remember the smallest sequence number that we use, so that we can
        // later identify whether a response belongs to the last search request.
        sequenceNumberAtBegin = resolver.currentSequenceNumber();
        // This is a new try, so we reset the searchRequestSent and
        // searchResponseReceived flags.
        searchRequestSent = false;
        searchResponseReceived = false;
        int numberOfFramesSent = 0;
        ChannelSearchRequest request;
        while ((request = firstValueFromMap(requestPending)) != null) {
            // We try to add the next search request. If this fails, because
            // the current packet is full, we send the packet and start with a
            // new one.
            AddSearchResult addResult = addSearchMessage(request);
            if (addResult == AddSearchResult.PACKET_FULL) {
                resolver.send();
                numberOfFramesSent++;
                if (((double) numberOfFramesSent) < framesPerTry) {
                    // We try to add the search request again.
                    continue;
                } else {
                    // The maximum number of frames has been reached, thus we
                    // quit the loop and process the request on the next run.
                    break;
                }
            } else if (addResult == AddSearchResult.EXCEPTION) {
                // The request has already been removed by
                // addSearchMessage(...), so we can simply continue with the
                // next request.
                continue;
            }
            // We added the request successfully (otherwise we would not be
            // here).
            requestPending.remove(request.getCID());
            responsePending.put(request.getCID(), request);
            // We added at least one search request, so we can set the flag.
            searchRequestSent = true;
        }
        // We have to send the last packet. If this packet is empty, calling
        // send() will simply do nothing.
        resolver.send();
        numberOfFramesSent++;
        // We remember the largest sequence number that we use, so that we can
        // later identify whether a response belongs to the last search request.
        sequenceNumberAtEnd = resolver.currentSequenceNumber();
        if (sequenceNumberAtEnd == Integer.MIN_VALUE) {
            sequenceNumberAtEnd = Integer.MAX_VALUE;
        } else {
            sequenceNumberAtEnd--;
        }
        // We schedule the next run for this timer.
        // TODO At the moment this is implemented exactly like in the C library.
        // However, it seems reasonable to use a shorter delay if we did not
        // process all requests.
        return System.currentTimeMillis() + intervalFactor
                * (long) Math.floor(rttEstimator.getEstimate());
    }

    /**
     * Returns the first value from a map without removing it.
     * 
     * @param map
     *            map from which the value is retrieved.
     * @return first value in the map or <code>null</code> if the map is empty.
     */
    private static <T> T firstValueFromMap(Map<?, ? extends T> map) {
        Iterator<? extends T> i = map.values().iterator();
        if (i.hasNext()) {
            return i.next();
        } else {
            return null;
        }
    }

    /**
     * Notifies this timer that a search request it sent was successful. This
     * method is called by {@link DefaultChannelNameResolverImpl} each time it
     * receives a search response for a request that was sent by this timer. If
     * multiple responses for the same request are received, only the first
     * response triggers this method. The timer uses this information to update
     * the round-trip-time estimate and to get feedback on the success rate of
     * the search requests it sends.
     * 
     * @param sequenceNumberIsValid
     *            <code>true</code> if the response was received via UDP and the
     *            corresponding flag was set. This means that the
     *            <code>sequenceNumber</code> can be used to identify whether
     *            the search request that was successful was sent during the
     *            last run of this timer. If <code>false</code>, the
     *            <code>sequenceNumber</code> should be ignored.
     * @param sequenceNumber
     *            UDP packet sequence number of the search request that was
     *            successful or an arbitrary number if
     *            <code>sequenceNumberIsValid</code> is <code>false</code>.
     * @return <code>true</code> if this timer detected a high success rate and
     *         thus its {@link #process()} method should be called as soon as
     *         possible, not waiting for the next interval.
     */
    public boolean processChannelFound(boolean sequenceNumberIsValid,
            int sequenceNumber) {
        if (sequenceNumberIsValid
                && (sequenceNumber < sequenceNumberAtBegin || sequenceNumber > sequenceNumberAtEnd)) {
            // The sequence number is valid but it is not in the range of
            // sequence numbers that we used in the last run. Therefore, it must
            // be an answer to an old search request.
            return false;
        }
        // The sequence number belongs to the last batch of search requests or
        // we cannot identify the batch (most likely because the message was
        // sent over TCP). In both cases, we update the round-trip-time
        // estimate.
        long measuredTime = System.currentTimeMillis() - timeAtLastSend;
        rttEstimator.updateEstimate(measuredTime);
        // We received at least one response. This information is used during
        // the next run of process().
        searchResponseReceived = true;
        if (responsePending.size() == 0) {
            // We reached a success rate of 100 percent. This means that
            // we should send all other search requests as soon as possible.
            return true;
        }
        return false;
    }

    private AddSearchResult addSearchMessage(ChannelSearchRequest request) {
        try {
            if (resolver.addSearchMessage(request.getCID(),
                    request.getChannelName())) {
                return AddSearchResult.SUCCESS;
            } else {
                return AddSearchResult.PACKET_FULL;
            }
        } catch (Exception e) {
            // We got an exception. We remove the request but we do not
            // add it to the list of pending responses. Effectively, this
            // means that the request will never be sent again. This makes
            // sense, because the reason for the exception must be in the
            // request.
            requestPending.remove(request.getCID());
            resolver.processFailedRequest(request, e);
            return AddSearchResult.EXCEPTION;
        }
    }

}
