/*
 * Copyright 2015-2019 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.resolver.internal;

import java.io.IOException;
import java.net.DatagramSocket;
import java.net.Inet4Address;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.net.SocketException;
import java.net.StandardProtocolFamily;
import java.net.StandardSocketOptions;
import java.nio.BufferOverflowException;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.DatagramChannel;
import java.nio.channels.SelectionKey;
import java.util.Collection;

import org.apache.commons.lang3.StringEscapeUtils;

import com.aquenos.epics.jackie.client.resolver.ChannelNameResolverConfiguration;
import com.aquenos.epics.jackie.common.exception.MalformedMessageException;
import com.aquenos.epics.jackie.common.exception.ReceivedMessageTooLargeException;
import com.aquenos.epics.jackie.common.exception.SerializedChannelNameTooLargeException;
import com.aquenos.epics.jackie.common.exception.SerializedPayloadTooLargeException;
import com.aquenos.epics.jackie.common.exception.UnsupportedMessageTypeException;
import com.aquenos.epics.jackie.common.io.ByteBufferByteSink;
import com.aquenos.epics.jackie.common.io.ByteBufferByteSource;
import com.aquenos.epics.jackie.common.io.ChannelProcessor;
import com.aquenos.epics.jackie.common.io.CommunicationController;
import com.aquenos.epics.jackie.common.io.CommunicationProcessor;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConstants;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessageCodec;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessSearchClientMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessSearchUDPServerMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersionMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersionUDPMessage;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;

/**
 * <p>
 * Channel name resolver for the UDP-based protocol.
 * </p>
 * 
 * <p>
 * This resolver uses a non-blocking {@link DatagramChannel} for sending and
 * receiving UDP packets. Received packets are processed immediately. Sent
 * packets might be queued by {@link UdpSender}. The {@link #pendingWrite()}
 * method can be used to check whether there are packets that have been queued
 * but not sent yet.
 * </p>
 * 
 * <p>
 * This implementation limits the size of sent packets to a maximum of 1024
 * bytes in order to stay compatible with all Channel Access servers. However,
 * it will receive larger packets (depending on the receive buffer size of the
 * underlying UDP socket).
 * </p>
 * 
 * <p>
 * This class is only intended for use by {@link DefaultChannelNameResolverImpl}
 * . It is closely tied to that class and has only been separated into a class
 * of its own to keep that class smaller.
 * </p>
 * 
 * <p>
 * This class is not thread safe, but it is only called by its
 * {@link CommunicationController} or methods of other classes that are managed
 * by the same {@link CommunicationController}.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class UdpResolver implements ChannelProcessor, CommunicationProcessor {

    /**
     * Maximum size of a sent UDP packet (in bytes). This limit is mandated by
     * the Channel Access protocol because old servers might not be able to deal
     * with larger packets.
     */
    private final static int MAXIMUM_PACKET_SEND_SIZE = 1024;

    /**
     * Desired size of the sockets send buffer (in bytes). Having a larger send
     * buffer allows us to send multiple packets (e.g. to different target
     * addresses) at once.
     */
    private final static int DESIRED_MIN_SOCKET_SEND_BUFFER_SIZE = 32768;

    /**
     * Desired size of the socket's receive buffer (in bytes). A larger receive
     * buffer makes it less likely that we lose packets because of the buffer
     * filling up before we can read them.
     */
    private final static int DESIRED_MIN_SOCKET_RECEIVE_BUFFER_SIZE = 32768;

    /**
     * Size of the buffer that we use for reading packets (in bytes). This
     * effectively limits the size of the largest packet that we can receive.
     * This limit is taken from the C library and should be sufficient to
     * receive a large number of messages in a single packet.
     */
    private final static int INTERNAL_RECEIVE_BUFFER_SIZE = 65551;

    /**
     * Maximum send size for a packet (in bytes). This is never greater than
     * {@link #MAXIMUM_PACKET_SEND_SIZE}, but it might be smaller if the UDP
     * socket's send buffer is smaller.
     */
    private final int maximumPacketSize;

    /**
     * Maximum size of a channel name after being serialized (including the
     * terminating null-byte). This limit is either
     * {@link ChannelAccessConstants#MAX_CHANNEL_NAME_BYTES_LENGTH}, or defines
     * by the maximum packet size, whichever is smaller.
     */
    private final int maximumSerializedChannelNameSize;

    /**
     * Message codec used for encoding and decoding messages that are sent and
     * received.
     */
    private ChannelAccessMessageCodec codec;

    /**
     * Channel that is used for sending and receiving UDP packets.
     */
    private DatagramChannel channel;

    /**
     * Selection key for {@link #channel}. The selection key is to register for
     * notifications when the channel is ready for reading or writing.
     */
    private SelectionKey channelSelectionKey;

    /**
     * Communication controller that is managing this object. The communication
     * controller supervises the channel and notifies this resolver when the
     * channel is ready for an I/O operation.
     */
    private CommunicationController communicationController;

    /**
     * Channel name resolver configuration.
     */
    private ChannelNameResolverConfiguration configuration;

    /**
     * Sequence number of the next packet. This is the sequence number of the
     * packet that will be created after the current packet has been filled up
     * and sent.
     */
    private int nextSequenceNumber;

    /**
     * Buffer used for receiving packets. For receiving, only a single buffer is
     * used because its content is decoded immediately.
     */
    private ByteBuffer receiveBuffer;

    /**
     * Byte source that is used for receiving packets. This is just used as a
     * wrapper around {@link #receiveBuffer} so that it can be handles by the
     * {@link #codec}.
     */
    private ByteBufferByteSource receiveSource;

    /**
     * Sequence number of the UDP packet that has just been received. The
     * sequence number is initialized as soon as a CA_PROTO_VERSION message is
     * found in the packet so that subsequent messages can use the sequence
     * number.
     */
    private int receivedSequenceNumber;

    /**
     * Is the {@link #receivedSequenceNumber} valid? This flag is initialized
     * with <code>false</code> every time a new packet is received. As soon as a
     * CA_PROTO_VERSION message with the sequence-number valid flag is found in
     * the received packet, this flag is set to <code>true</code>.
     */
    private boolean receivedSequenceNumberValid;

    /**
     * The channel name resolver that created this UDP resolver.
     */
    private DefaultChannelNameResolverImpl resolver;

    /**
     * Byte sink used for creating packets to be sent. The byte sink is
     * initialized with a single byte buffer by {@link #prepareSendSink()}. When
     * this buffer is full, {@link #send()} has to be called which removes the
     * full buffer, queues it for sending and finally calls
     * {@link #prepareSendSink()} to add a new buffer to the byte sink.
     */
    private ByteBufferByteSink sendSink;

    /**
     * Have search messages been added to the {@link #sendSink}? Each time the
     * send sink is initialized with a new buffer, this flag is reset to
     * <code>false</code>. It is set to <code>true</code> the first time a
     * search message is added to this buffer.
     */
    private boolean sendSinkDirty;

    /**
     * UDP sender that is internally used to send UDP packets. The sender takes
     * care of queueing packets until they can be sent.
     */
    private UdpSender sender;

    /**
     * Creates a UDP resolver that is owned by <code>resolver</code>.
     * 
     * @param configuration
     *            channel-name resolver configuration.
     * @param resolver
     *            channel name resolver owning this resolver.
     * @throws IOException
     *             if the channel that is needed for sending and receiving UDP
     *             packets or its corresponding socket cannot be setup.
     */
    public UdpResolver(ChannelNameResolverConfiguration configuration,
            DefaultChannelNameResolverImpl resolver) throws IOException {
        this.configuration = configuration;
        this.resolver = resolver;
        boolean successful = false;
        this.channel = DatagramChannel.open(StandardProtocolFamily.INET);
        // Bind the socket to a random port and all interfaces
        try {
            DatagramSocket socket = this.channel.socket();
            this.channel.socket().bind(new InetSocketAddress(0));
            try {
                // Setting the broadcast flag is not critical, because it will
                // work without broadcasts, if unicast addresses are specified
                // explicitly. Therefore, we do not fail if we cannot get the
                // broadcast capability.
                socket.setBroadcast(true);
            } catch (SocketException e) {
                configuration.getErrorHandler().handleError(this.getClass(), e,
                        "Cannot enable broadcasts on UDP socket for name resolution.");
            }
            // We try to increase the socket buffer sizes if necessary. If this
            // fails, it is a non-critical error, because communication might
            // still work. However, we make sure that we do not send packets
            // that are larger than the send buffer, because depending on the
            // network-stack implementation this might fail.
            if (socket
                    .getReceiveBufferSize() < DESIRED_MIN_SOCKET_RECEIVE_BUFFER_SIZE) {
                try {
                    socket.setReceiveBufferSize(
                            DESIRED_MIN_SOCKET_RECEIVE_BUFFER_SIZE);
                } catch (SocketException e) {
                    configuration.getErrorHandler().handleError(this.getClass(),
                            e,
                            "Cannot increase receive buffer size on UDP socket for name resolution.");
                }
            }
            // We try to increase the send buffer size of the socket. Having a
            // larger send buffer allows us to send multiple packets at once.
            if (socket
                    .getSendBufferSize() < DESIRED_MIN_SOCKET_SEND_BUFFER_SIZE) {
                try {
                    socket.setSendBufferSize(
                            DESIRED_MIN_SOCKET_SEND_BUFFER_SIZE);
                } catch (SocketException e) {
                    configuration.getErrorHandler().handleError(this.getClass(),
                            e,
                            "Cannot increase send buffer size on UDP socket for name resolution.");
                }
            }
            // If we cannot have a send buffer that is as large as we would
            // like, we try to make it at least as large as the maximum size of
            // a single packet.
            if (socket.getSendBufferSize() < MAXIMUM_PACKET_SEND_SIZE) {
                try {
                    socket.setSendBufferSize(MAXIMUM_PACKET_SEND_SIZE);
                } catch (SocketException e) {
                    configuration.getErrorHandler().handleError(this.getClass(),
                            e,
                            "Cannot increase send buffer size on UDP socket for name resolution.");
                }
            }
            // We have to limit the packet size to the size of the send buffer
            // or the maximum size allowed by the Channel Access protocol,
            // whichever is smaller.
            this.maximumPacketSize = Math.min(socket.getSendBufferSize(),
                    MAXIMUM_PACKET_SEND_SIZE);
            if (this.maximumPacketSize < (2
                    * ChannelAccessConstants.STANDARD_HEADER_SIZE + 40)) {
                // We need space for at least two (small) headers
                // (CA_PROTO_VERSION + CA_PROTO_SEARCH) and the payload of the
                // CA_PROTO_SEARCH message (40 seems to be a reasonable minimum
                // because this is a typical limit for channel names).
                throw new SocketException(
                        "The socket's send buffer size is too small and could not be increased.");
            }
            // We try to set both the IP_MULTICAST_LOOP and IP_MULTICAST_TTL
            // options. This might fail due to platform restrictions. In this
            // case, we will simply operate without these options, because they
            // should not have any effect when sending unicast or broadcast
            // packets (and multicast packets are only sent in very specific
            // configurations).
            try {
                channel.setOption(StandardSocketOptions.IP_MULTICAST_LOOP,
                        true);
            } catch (UnsupportedOperationException e) {
                // If the platform we are running on does not support the
                // IP_MULTICAST_LOOP option, we still want to continue.
                configuration.getErrorHandler().handleError(getClass(), e,
                        "IP_MULTICAST_LOOP option could not be set on UDP socket for name resolution.");
            }
            try {
                channel.setOption(StandardSocketOptions.IP_MULTICAST_TTL,
                        configuration.getMulticastTimeToLive());
            } catch (UnsupportedOperationException e) {
                // If the platform we are running on does not support the
                // IP_MULTICAST_TTL option, we still want to continue.
                configuration.getErrorHandler().handleError(getClass(), e,
                        "IP_MULTICAST_TTL option could not be set on UDP socket for name resolution.");
            }
            // Calculate the maximum size of a channel name. If the maximum
            // packet size is rather limited, this might be smaller than the
            // default limit of 488 bytes. The calculated limit might not be
            // aligned to 8 bytes, but this is okay because the check for the
            // channel name length takes care of alignment.
            this.maximumSerializedChannelNameSize = Math.min(
                    ChannelAccessConstants.MAX_CHANNEL_NAME_BYTES_LENGTH,
                    maximumPacketSize
                            - 2 * ChannelAccessConstants.STANDARD_HEADER_SIZE);
            // Set channel to non-blocking mode. This is critical, because we
            // cannot use selectors with a blocking channel.
            this.channel.configureBlocking(false);
            // We use direct buffers for sending and receiving data, because
            // this helps us avoid unnecessary copy operations.
            this.receiveBuffer = ByteBuffer
                    .allocateDirect(INTERNAL_RECEIVE_BUFFER_SIZE);
            successful = true;
        } finally {
            if (!successful) {
                try {
                    channel.close();
                } catch (IOException e) {
                    // We ignore any exception that might be thrown while
                    // closing the channel.
                }
            }
        }
        this.codec = new ChannelAccessMessageCodec();
        this.codec.setCharset(configuration.getCharset());
        this.nextSequenceNumber = 1;
        this.receiveSource = new ByteBufferByteSource();
        this.sendSink = new ByteBufferByteSink();
        this.sender = new UdpSender(this.channel);
        prepareSendSink();
    }

    /**
     * Allocates a byte buffer that can be used for sending a packet. This
     * buffer always has a size of {@link #maximumPacketSize}. This method tries
     * to get a free buffer from the free list maintained by {@link #sender} and
     * allocates a new buffer if this free list is empty. Buffers that are
     * allocated by this method are always direct.
     * 
     * @return byte buffer with a size of {@link #maximumPacketSize}, ready for
     *         writing.
     */
    private ByteBuffer allocateSendBuffer() {
        ByteBuffer sendBuffer = sender.nextFreeBuffer();
        if (sendBuffer != null) {
            return sendBuffer;
        } else {
            // The send buffer size is chosen so that it is equal to the maximum
            // packet size. This way, each packet that we can sent fits into the
            // buffer and the buffer always fits into a packet.
            return ByteBuffer.allocateDirect(maximumPacketSize);
        }
    }

    /**
     * Prepares the {@link #sendSink} so that search messages can be added. This
     * means that a send buffer (allocated by {@link #allocateSendBuffer()}) is
     * added to the byte sink, a CA_PROTO_VERSION message is written to the byte
     * sink and the {@link #nextSequenceNumber} is incremented.
     */
    private void prepareSendSink() {
        sendSink.addBuffer(allocateSendBuffer());
        sendSinkDirty = false;
        ChannelAccessVersionMessage versionMessage = new ChannelAccessVersionUDPMessage(
                ChannelAccessVersion.NEWEST_SUPPORTED_VERSION, true,
                nextSequenceNumber);
        if (nextSequenceNumber == Integer.MAX_VALUE) {
            nextSequenceNumber = Integer.MIN_VALUE;
        } else {
            nextSequenceNumber++;
        }
        codec.encodeMessageToUDPServer(sendSink, versionMessage,
                ChannelAccessVersion.NEWEST_SUPPORTED_VERSION);
    }

    /**
     * Adds a search message to the current packet buffer. If the space
     * remaining in the buffer is too small to add the complete message, the
     * message is not added <code>false</code> is returned. If the space
     * remaining in the buffer is too small even though this is the first search
     * message being added, a {@link SerializedChannelNameTooLargeException} is
     * thrown.
     * 
     * @param cid
     *            numeric, client-assigned channel ID. This ID is used to
     *            identify the channel in a response to the search request.
     * @param channelName
     *            channel name.
     * @return <code>true</code> if the search message has been added to the
     *         buffer for the current UDP packet, <code>false</code> if it
     *         cannot be added because the space remaining in the buffer is
     *         insufficient.
     * @throws SerializedChannelNameTooLargeException
     *             if the space remaining in the packet buffer is insufficient
     *             even though this is the first search message being added or
     *             if the serialized form of the channel name exceed a general
     *             limit (
     *             {@link ChannelAccessConstants#MAX_CHANNEL_NAME_BYTES_LENGTH}
     *             ).
     */
    public boolean addSearchMessage(int cid, String channelName) {
        ChannelAccessSearchClientMessage searchMessage = new ChannelAccessSearchClientMessage(
                cid, ChannelAccessVersion.NEWEST_SUPPORTED_VERSION, false,
                channelName);
        try {
            codec.encodeMessageToUDPServer(sendSink, searchMessage,
                    ChannelAccessVersion.NEWEST_SUPPORTED_VERSION);
        } catch (BufferOverflowException e) {
            if (!sendSinkDirty) {
                // This was the first message that we tried to add and it still
                // did not fit into the buffer. This means that the message is
                // too large (the channel name is too long) and thus we will
                // never be able to send it.
                throw new SerializedChannelNameTooLargeException(
                        "Channel name \""
                                + StringEscapeUtils.escapeJava(channelName)
                                + "\" is too large to be serialized into the UDP send buffer.");
            }
            return false;
        } catch (SerializedPayloadTooLargeException e) {
            // The only payload is the channel name. We replace the exception so
            // that we only have one type.
            throw new SerializedChannelNameTooLargeException("Channel name \""
                    + StringEscapeUtils.escapeJava(channelName)
                    + "\" is too large to be serialized into the UDP send buffer.",
                    e);
        }
        sendSinkDirty = true;
        return true;
    }

    /**
     * Returns the sequence number of the packet that is currently being
     * assembled. This is the packet that will be sent when {@link #send()} is
     * called.
     * 
     * @return sequence number of the current UDP packet.
     */
    public int currentSequenceNumber() {
        if (nextSequenceNumber == Integer.MIN_VALUE) {
            return Integer.MAX_VALUE;
        } else {
            return nextSequenceNumber - 1;
        }
    }

    /**
     * Sends the current UDP packet. This method should be called to actually
     * send the search messages that have been added by calling
     * {@link #addSearchMessage(int, String)}.
     */
    public void send() {
        if (!sendSinkDirty) {
            // If we have not added any search messages yet, we do not want to
            // send the packet.
            return;
        }
        // We know that the sendSink always has exactly one buffer, thus we can
        // keep this simple.
        ByteBuffer sendBuffer = sendSink.getWrittenData()[0];
        sender.sendBuffer(sendBuffer);
        // We try to actually send the packets immediately. If the socket's
        // send buffer is sufficiently large, this will actually succeed.
        processWrite();
        // Now we prepare the send sink for the next packet. We do not do this
        // earlier because this way, we can directly reuse the buffer, if
        // processWrite could completely write all data.
        prepareSendSink();
    }

    public void setTargetAddresses(
            Collection<? extends InetSocketAddress> targetAddresses) {
        sender.setTargetAddresses(targetAddresses);
    }

    /**
     * Verifies whether a channel name can be sent in a UDP search request.
     * 
     * @param channelName
     *            channel name to be verified.
     * @throws SerializedChannelNameTooLargeException
     *             if the channel name is too large to be sent.
     */
    public void verifyChannelName(String channelName) {
        if (NullTerminatedStringUtil
                .stringToSizeLimitedNullTerminatedBytesOrNull(channelName,
                        maximumSerializedChannelNameSize,
                        ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT,
                        codec.getCharset()) == null) {
            throw new SerializedChannelNameTooLargeException(
                    "Channel name too large to be serialized: " + channelName);
        }
    }

    /**
     * Destroys this resolver. This closes the channel used for sending and
     * receiving UDP packets. Subsequent I/O operations will fail with an
     * exception.
     */
    public void destroy() {
        try {
            channel.close();
        } catch (IOException e) {
            // We ignore any exception that might be thrown while
            // closing the channel.
        }
    }

    /**
     * Tells whether there is data that is waiting to be sent.
     * 
     * @return <code>true</code> if packets have been queued but not completely
     *         sent yet, <code>false</code> if the send queue is empty.
     */
    public boolean pendingWrite() {
        return sender.pendingWrite();
    }

    @Override
    public void processIO() {
        if (channelSelectionKey.isReadable()) {
            processRead();
        }
        if (channelSelectionKey.isWritable()) {
            processWrite();
        }
    }

    /**
     * Reads from the {@link #channel} until there are no packets left.
     */
    private void processRead() {
        SocketAddress sourceAddress;
        do {
            receiveBuffer.clear();
            try {
                sourceAddress = channel.receive(receiveBuffer);
            } catch (ClosedChannelException e) {
                // If the channel is closed there is no sense in continuing.
                throw new IllegalStateException(e);
            } catch (IOException e) {
                // We log the exception and do not continue because most likely,
                // the receive operation would just throw another exception.
                configuration.getErrorHandler().handleError(this.getClass(), e,
                        "Error while trying to receive a UDP packet.");
                return;
            }
            // The source address should always be an InetSocketAddress storing
            // an Inet4Address, but if it is not, we simply ignore the packet.
            if (sourceAddress != null
                    && sourceAddress instanceof InetSocketAddress) {
                InetSocketAddress inetSourceAddress = (InetSocketAddress) sourceAddress;
                if (inetSourceAddress.getAddress() instanceof Inet4Address) {
                    Inet4Address inet4Address = (Inet4Address) inetSourceAddress
                            .getAddress();
                    receiveBuffer.flip();
                    receiveSource.appendBuffer(receiveBuffer);
                    processReceivedPacket(inet4Address,
                            inetSourceAddress.getPort());
                }
            }
        } while (sourceAddress != null);
    }

    /**
     * Processes a packet that has been received. Decodes all messages contained
     * in this packet and handles them.
     */
    private void processReceivedPacket(Inet4Address sourceAddress,
            int sourcePort) {
        receivedSequenceNumberValid = false;
        while (receiveSource.remaining() > 0) {
            ChannelAccessMessage message;
            try {
                try {
                    message = codec.decodeMessageToUDPClient(receiveSource,
                            ChannelAccessVersion.NEWEST_SUPPORTED_VERSION,
                            sourceAddress, sourcePort);
                } catch (MalformedMessageException e) {
                    configuration.getErrorHandler().handleError(
                            UdpResolver.class, e,
                            "Received a malformed message via UDP.");
                    codec.discardNextMessage(receiveSource);
                    continue;
                } catch (ReceivedMessageTooLargeException e) {
                    configuration.getErrorHandler().handleError(
                            UdpResolver.class, e,
                            "Received a too large message via UDP.");
                    codec.discardNextMessage(receiveSource);
                    continue;
                } catch (UnsupportedMessageTypeException e) {
                    configuration.getErrorHandler().handleError(
                            UdpResolver.class, e,
                            "Received a message of an unsupported type via UDP.");
                    codec.discardNextMessage(receiveSource);
                    continue;
                }
            } catch (BufferUnderflowException e) {
                configuration.getErrorHandler().handleError(UdpResolver.class,
                        null, "Ignoring extra bytes in UDP packet.");
                // Drain all data so that the receive buffer is released and can
                // be reused.
                receiveSource.skip(receiveSource.remaining());
                continue;
            } catch (Exception e) {
                configuration.getErrorHandler().handleError(UdpResolver.class,
                        e,
                        "Unexpected exception while trying to read a UDP packet.");
                // Drain all data so that the receive buffer is released and can
                // be reused.
                receiveSource.skip(receiveSource.remaining());
                continue;
            }
            processReceivedMessage(message);
        }
    }

    /**
     * Processes a message that has been received.
     * 
     * @param message
     *            the message that has been received. Must be of type
     *            CA_PROTO_VERSION or CA_PROTO_SEARCH.
     */
    private void processReceivedMessage(ChannelAccessMessage message) {
        switch (message.getCommand()) {
        case CA_PROTO_VERSION:
            processVersionMessage((ChannelAccessVersionUDPMessage) message);
            break;
        case CA_PROTO_SEARCH:
            processSearchMessage((ChannelAccessSearchUDPServerMessage) message);
            break;
        default:
            configuration.getErrorHandler().handleError(UdpResolver.class, null,
                    "Received unexpected message via UDP: "
                            + message.toString());
        }
    }

    /**
     * Processes a CA_PROTO_VERSION message that has been received.
     * 
     * @param message
     *            version message.
     */
    private void processVersionMessage(ChannelAccessVersionUDPMessage message) {
        if (!message.isSequenceNumberValid()) {
            return;
        }
        receivedSequenceNumberValid = true;
        receivedSequenceNumber = message.getSequenceNumber();
    }

    /**
     * Processes a CA_PROTO_SEARCH message that has been received.
     * 
     * @param message
     *            search (response) message.
     */
    private void processSearchMessage(
            ChannelAccessSearchUDPServerMessage message) {
        resolver.processSearchResponse(message, receivedSequenceNumberValid,
                receivedSequenceNumber);
    }

    /**
     * Writes data from the send queue to the channel. Notifies the channel
     * resolver when all data has been written so that it can run tasks that
     * have been waiting for the send queue to be cleared.
     */
    private void processWrite() {
        boolean pendingWriteBefore = sender.pendingWrite();
        try {
            sender.processWrite();
        } catch (ClosedChannelException e) {
            // If the channel is closed there is no sense in trying with
            // other addresses.
            throw new IllegalStateException(e);
        } catch (Exception e) {
            // We log the exception and then continue with the next
            // address.
            configuration.getErrorHandler().handleError(this.getClass(), e,
                    "Error while trying to send a UDP packet.");
        }
        if (pendingWriteBefore && !sender.pendingWrite()) {
            // We just finished writing all pending requests. This means that
            // we should schedule a timer event in the channel-name resolver
            // so that it can process timers that might have been put on hold.
            resolver.scheduleImmediateTimerProcessing();
        }
    }

    @Override
    public void setCommunicationController(
            CommunicationController communicationController) {
        this.communicationController = communicationController;
        try {
            this.channelSelectionKey = this.communicationController
                    .registerChannel(this, channel);
        } catch (ClosedChannelException e) {
            // This can only happen if this object has already been destroyed.
            throw new IllegalStateException(e);
        }
        // We are always interested in read-events (received packets).
        this.channelSelectionKey.interestOps(SelectionKey.OP_READ);
        this.sender.setSelectionKey(this.channelSelectionKey);
    }

}
