/*
 * Copyright 2014-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.io;

import java.nio.BufferOverflowException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.Channel;
import java.util.LinkedList;
import java.util.ListIterator;

/**
 * <p>
 * Byte sink that is backed by a list of byte buffers. This byte sink allows for
 * accessing a sequence of buffers like they were one contiguous buffer.
 * </p>
 * 
 * <p>
 * <strong>Attention</strong>: Special care must be taken to ensure that all
 * byte buffers added to an instance of this class have the same
 * {@link ByteOrder}. Otherwise the put methods that return multi-byte
 * data-types might have unexpected results.
 * </p>
 * 
 * <p>
 * The <code>put*Array</code> methods are optimized to only use one copy
 * operation for transferring data. An extra copy operation is necessary for
 * elements that span across two of the backing {@link ByteBuffer}s, however all
 * other elements are written with only a single copy operation. If the
 * {@link ByteBuffer}'s {@link ByteOrder} is the platforms native byte order,
 * array writes can be expected to be very efficient, because complete blocks of
 * memory are copied.
 * </p>
 * 
 * <p>
 * Sub-classes can override {@link #requestAdditionalBuffer(int)} in order to
 * provide an additional buffer if this byte sink runs out of space.
 * </p>
 * 
 * <p>
 * Instances of this class are not thread-safe. If multiple threads are going to
 * access an instance of this class, access to it must be synchronized. This
 * also applies to both simple put operations as well as "atomic" put
 * operations.
 * </p>
 * 
 * <p>
 * This implementation guarantees that if a put operation finished successfully,
 * the whole data will have been written to the backing byte-buffers.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class ByteBufferByteSink extends AbstractByteSink {

    /**
     * Interface for a command that copies data from an array of a specific type
     * into a byte buffer. This would be a perfect example for lambda
     * expressions, but we do not want to require Java 8 yet.
     * 
     * @author Sebastian Marsching
     * 
     * @param <T>
     *            type of the array from which this reader can write data.
     */
    private static interface ValueWriter<T> {

        /**
         * Writes elements from <code>array</code> into <code>buffer</code>.
         * This method may change the position of the byte buffer or other
         * properties. Obviously, it also changes the buffer's content.
         * 
         * @param source
         *            array from which the elements should be copied.
         * @param sourceOffset
         *            offset into <code>source</code>. The method should start
         *            reading elements at this position.
         * @param destination
         *            buffer to which the elements should be copied. This method
         *            may change this buffer's properties (e.g. the position).
         * @param numberOfBytes
         *            number of bytes that should be written. The calling code
         *            ensures that there are always enough remaining bytes in
         *            the buffer in order to write the specified number of
         *            elements and that the array is big enough to read the
         *            specified number of elements (also considering the
         *            <code>sourceOffset</code>).
         */
        void writeValues(T source, int sourceOffset, ByteBuffer destination,
                int numberOfElements);

    }

    private final static ValueWriter<byte[]> BYTE_WRITER = new ValueWriter<byte[]>() {
        @Override
        public void writeValues(byte[] source, int sourceOffset,
                ByteBuffer destination, int numberOfBytes) {
            destination.put(source, sourceOffset, numberOfBytes);
        }
    };

    private final static ValueWriter<double[]> DOUBLE_WRITER = new ValueWriter<double[]>() {
        @Override
        public void writeValues(double[] source, int sourceOffset,
                ByteBuffer destination, int numberOfBytes) {
            assert (numberOfBytes % SIZE_OF_DOUBLE == 0);
            int numberOfElements = numberOfBytes / SIZE_OF_DOUBLE;
            destination.asDoubleBuffer().put(source, sourceOffset,
                    numberOfElements);
        }
    };

    private final static ValueWriter<float[]> FLOAT_WRITER = new ValueWriter<float[]>() {
        @Override
        public void writeValues(float[] source, int sourceOffset,
                ByteBuffer destination, int numberOfBytes) {
            assert (numberOfBytes % SIZE_OF_FLOAT == 0);
            int numberOfElements = numberOfBytes / SIZE_OF_FLOAT;
            destination.asFloatBuffer().put(source, sourceOffset,
                    numberOfElements);
        }
    };

    private final static ValueWriter<int[]> INT_WRITER = new ValueWriter<int[]>() {
        @Override
        public void writeValues(int[] source, int sourceOffset,
                ByteBuffer destination, int numberOfBytes) {
            assert (numberOfBytes % SIZE_OF_INT == 0);
            int numberOfElements = numberOfBytes / SIZE_OF_INT;
            destination.asIntBuffer().put(source, sourceOffset,
                    numberOfElements);
        }
    };

    private final static ValueWriter<short[]> SHORT_WRITER = new ValueWriter<short[]>() {
        @Override
        public void writeValues(short[] source, int sourceOffset,
                ByteBuffer destination, int numberOfBytes) {
            assert (numberOfBytes % SIZE_OF_SHORT == 0);
            int numberOfElements = numberOfBytes / SIZE_OF_SHORT;
            destination.asShortBuffer().put(source, sourceOffset,
                    numberOfElements);
        }
    };

    private final static ValueWriter<long[]> LONG_WRITER = new ValueWriter<long[]>() {
        @Override
        public void writeValues(long[] source, int sourceOffset,
                ByteBuffer destination, int numberOfBytes) {
            assert (numberOfBytes % SIZE_OF_LONG == 0);
            int numberOfElements = numberOfBytes / SIZE_OF_LONG;
            destination.asLongBuffer().put(source, sourceOffset,
                    numberOfElements);
        }
    };

    private LinkedList<ByteBuffer> nextBuffers = new LinkedList<ByteBuffer>();
    private LinkedList<ByteBuffer> fullBuffers = new LinkedList<ByteBuffer>();

    private boolean nextBufferIsDirty = false;

    private boolean atomicWriteInProgress = false;
    private long atomicWriteOffset = 0;
    private LinkedList<Long> atomicWriteOffsetStack = new LinkedList<Long>();

    private long bytesWritten = 0L;

    /**
     * Creates a new byte sink that initially does not contain any buffers. This
     * means that the write methods throw a {@link BufferOverflowException}
     * until at least one usable buffer is added to this sink.
     */
    public ByteBufferByteSink() {
    }

    /**
     * Adds a byte buffer to the list of available byte-buffers. The buffer will
     * be used to store data written to this byte sink.
     * 
     * @param byteBuffer
     *            byte buffer that shall be used to store data.
     * @return reference to this byte sink.
     * @throws IllegalArgumentException
     *             if the byte buffer's position is not zero (there is already
     *             data in the byte buffer). The byte buffer is not added to the
     *             list of available buffers.
     */
    public ByteBufferByteSink addBuffer(ByteBuffer byteBuffer) {
        // We check whether the byte buffer's position is zero. This avoids a
        // situation where data could sneak in because a buffer has not been
        // cleared before making it available for write operations.
        if (byteBuffer.position() != 0) {
            throw new IllegalArgumentException(
                    "The byte buffer's position must be zero.");
        }
        nextBuffers.addLast(byteBuffer);
        return this;
    }

    /**
     * Adds byte buffers to the list of available byte-buffers. The buffers will
     * be used to store data written to this byte sink.
     * 
     * @param byteBuffers
     *            byte buffers that shall be used to store data.
     * @return reference to this byte sink.
     * @throws IllegalArgumentException
     *             if one of the byte buffers' positions is not zero (there is
     *             already data in the byte buffer). The byte buffers are not
     *             added to the list of available buffers.
     */
    public ByteBufferByteSink addBuffers(ByteBuffer[] byteBuffers) {
        boolean successful = false;
        int count = 0;
        try {
            for (ByteBuffer byteBuffer : byteBuffers) {
                addBuffer(byteBuffer);
                count++;
            }
            successful = true;
        } finally {
            // If the operation is not successful, we have to remove all buffers
            // that have already been added.
            if (!successful) {
                while (count > 0) {
                    nextBuffers.removeLast();
                    count--;
                }
            }
        }
        return this;
    }

    /**
     * Adds byte buffers to the list of available byte-buffers. The buffers will
     * be used to store data written to this byte sink.
     * 
     * @param byteBuffers
     *            byte buffers that shall be used to store data.
     * @return reference to this byte sink.
     * @throws IllegalArgumentException
     *             if one of the byte buffers' positions is not zero (there is
     *             already data in the byte buffer). The byte buffers are not
     *             added to the list of available buffers.
     */
    public ByteBufferByteSink addBuffers(
            Iterable<? extends ByteBuffer> byteBuffers) {
        boolean successful = false;
        int count = 0;
        try {
            for (ByteBuffer byteBuffer : byteBuffers) {
                addBuffer(byteBuffer);
                count++;
            }
            successful = true;
        } finally {
            // If the operation is not successful, we have to remove all buffers
            // that have already been added.
            if (!successful) {
                while (count > 0) {
                    nextBuffers.removeLast();
                    count--;
                }
            }
        }
        return this;
    }

    /**
     * <p>
     * Adds a byte buffer to the head of the list of available byte buffers.
     * This method may only be called when no atomic put operation is in
     * progress and no data is in the byte sink (no data has been written since
     * the last call to {@link #getWrittenData()}). The byte buffer that is
     * added may contain data. This data is treated like it had been written to
     * this byte sink. In particular, a call too
     * {@link #getWrittenDataSizeInBytes()} right after adding the buffer will
     * return a number equal to the buffer's position.
     * </p>
     * 
     * <p>
     * This method can be used to add a partially filled buffer that has been
     * removed using {@link #getWrittenData()} but could not be passed on (e.g.
     * because the channel it was supposed to be written to did not accept more
     * data). It is important that the buffer only contains data that should be
     * used, because all data from zero up to its current position is going to
     * be used. If the buffer contains data that should not be used, it should
     * first be prepared, e.g. by calling its {@link ByteBuffer#compact()
     * compact} method.
     * </p>
     * 
     * @param byteBuffer
     *            byte buffer that should be added to the list of available
     *            buffers before all other buffers (never <code>null</code>).
     *            The buffer may contain data, which is implicitly added to this
     *            byte sink.
     * @throws IllegalStateException
     *             if an atomic put operation is in progress or this byte sink
     *             already contains data.
     */
    public void addBufferFirst(ByteBuffer byteBuffer) {
        if (byteBuffer == null) {
            throw new NullPointerException();
        }
        // We do not allow to prepend a buffer while an atomic write is in
        // progress. If we allowed this, chances would be high that this would
        // mess up the internal state.
        if (atomicWriteInProgress) {
            throw new IllegalStateException(
                    "A buffer cannot be prepended while an atomic put operation is in progress.");
        }
        // We cannot prepend a buffer if data has already been written: This
        // would mess up the order of data.
        if (getWrittenDataSizeInBytes() != 0L) {
            throw new IllegalStateException(
                    "A buffer cannot be prepended when the byte sink is not empty (data has already been written to the byte sink but not removed).");
        }
        nextBuffers.addFirst(byteBuffer);
        if (byteBuffer.position() != 0) {
            nextBufferIsDirty = true;
            bytesWritten += byteBuffer.position();
        }
    }

    /**
     * Returns all byte buffers to which data has been written and removes them
     * from the internal list. This method may not be called from within an
     * atomic put operation.
     * 
     * @return array containing all byte buffers that have data. The buffers are
     *         returned in the correct order and have all been flipped, so that
     *         they can directly be used for a write to a {@link Channel}.
     * @throws IllegalStateException
     *             if an atomic put operation is in progress.
     */
    public ByteBuffer[] getWrittenData() {
        if (atomicWriteInProgress) {
            throw new IllegalStateException(
                    "This method must not be called while an atomic put operation is in progress.");
        }
        ByteBuffer[] result = new ByteBuffer[fullBuffers.size()
                + (nextBufferIsDirty ? 1 : 0)];
        int resultIndex = 0;
        for (ByteBuffer byteBuffer : fullBuffers) {
            // Prepare the byte buffer for reading.
            byteBuffer.flip();
            result[resultIndex] = byteBuffer;
            resultIndex++;
        }
        fullBuffers.clear();
        if (nextBufferIsDirty) {
            ByteBuffer nextBuffer = nextBuffers.pop();
            nextBuffer.flip();
            result[resultIndex] = nextBuffer;
            resultIndex++;
            nextBufferIsDirty = false;
        }
        bytesWritten = 0L;
        return result;
    }

    /**
     * Returns the number of bytes that have been written to this byte sink and
     * are stored in buffers that have not been removed yet. A call to
     * {@link #getWrittenData()} will return byte buffers that in total contain
     * this number of bytes.
     * 
     * @return number of bytes stored in the write buffers.
     */
    public long getWrittenDataSizeInBytes() {
        return bytesWritten + atomicWriteOffset;
    }

    /**
     * This method is called if the space remaining in the available
     * byte-buffers is not sufficient for a put operation. This method may
     * return an additional buffer (e.g. taken from a pool or newly allocated).
     * If this method cannot or does not want to provide an additional buffer,
     * it returns <code>null</code> in this case the write operation will fail
     * with an {@link BufferOverflowException}. The default implementation
     * always returns <code>null</code>, limiting the amount of data that can be
     * written to the space in the byte buffers that have been added to this
     * byte sink.
     * 
     * @param requiredSpaceInBytes
     *            number of bytes that are needed to complete the put operation.
     *            This information can be used to decide on the size that the
     *            buffer should have. However, the returned buffer may be larger
     *            or smaller than this size. If it is larger, the remaining
     *            space will be used for future put operations. If it is
     *            smaller, this method will be called again, requesting another
     *            buffer.
     * @return an additional byte-buffer that may be used to complete the write
     *         operation. If the return value is <code>null</code> no additional
     *         buffer is provided and the put operation in progress will fail
     *         with an {@link BufferOverflowException}. The default
     *         implementation always returns <code>null</code>.
     */
    protected ByteBuffer requestAdditionalBuffer(int requiredSpaceInBytes) {
        // By default we return null. This means that we only use the buffers
        // that have been added explicitly and never allocate a buffer
        // implicitly.
        return null;
    }

    /**
     * Writes the specified number of bytes from the supplied array, converting
     * them using the supplied value writer.
     * 
     * @param numberOfValues
     *            number of values to write.
     * @param elementSize
     *            size of each value (in bytes).
     * @param source
     *            array from which to copy the values to be written.
     * @param sourceOffset
     *            offset into the source array. The values are read from the
     *            array starting at this position.
     * @param valueWriter
     *            value writer for the specified array type. The value writer
     *            converts the values from the source type to bytes and puts
     *            them into the buffer.
     */
    private <T> void writeValues(int numberOfValues, int elementSize, T source,
            int sourceOffset, ValueWriter<T> valueWriter) {
        // If numberOfValues or sourceOffset is negative, this should already
        // have been caught, therefore we only use assertions for these checks
        // here.
        assert (numberOfValues >= 0);
        assert (sourceOffset >= 0);
        // We have to check that (considering the current offset) the number of
        // bytes we want to write does not become too large.
        long bytesToBeWritten = ((long) numberOfValues) * ((long) elementSize);
        if (Long.MAX_VALUE - atomicWriteOffset - bytesWritten < bytesToBeWritten) {
            // If we wrote the values, bytesWritten + atomicWriteOffset would
            // exceed Long.MAX_VALUE, thus we cannot allow this write. We treat
            // this as a buffer overflow (not enough space left).
            throw new BufferOverflowException();
        }
        // The element size should never come from an external parameter,
        // therefore we only use an assertion.
        assert (elementSize >= 1);

        writeValuesInternal(atomicWriteOffset, numberOfValues, elementSize,
                source, sourceOffset, valueWriter);

        if (atomicWriteInProgress) {
            atomicWriteOffset += bytesToBeWritten;
        } else {
            commit(bytesToBeWritten);
        }

    }

    private <T> void writeValuesInternal(long offset, int numberOfValues,
            int elementSize, T source, int sourceOffset,
            ValueWriter<T> valueWriter) {
        // We only make assertions on the input parameters. They should already
        // have been checked before.
        assert (offset >= 0);
        assert (numberOfValues >= 0);
        assert (elementSize >= 1);
        assert (sourceOffset >= 0);

        long remainingOffset = offset;
        // The product of numberOfValues and elementSize might be greater than
        // Integer.MAX_VALUE.
        long remainingLength = ((long) numberOfValues) * ((long) elementSize);

        ByteBuffer temporaryBuffer = null;

        ListIterator<ByteBuffer> nextBuffersIterator = nextBuffers
                .listIterator();
        ByteBuffer buffer;
        while ((buffer = bufferFromIteratorOrNewBuffer(nextBuffersIterator,
                remainingLength)) != null) {
            int remainingInBuffer = buffer.remaining();
            // We skip empty buffers.
            if (remainingInBuffer == 0) {
                continue;
            }
            // If we have a buffer that is smaller than the current offset, we
            // can skip the complete buffer and just decrease the offset.
            if (remainingInBuffer <= remainingOffset) {
                remainingOffset -= remainingInBuffer;
                continue;
            }
            int remainingUsableBytesInBuffer = remainingInBuffer
                    - (int) remainingOffset;
            // If the temporary buffer has not been initialized yet, we treat
            // this like an empty buffer (the buffer has no data).
            int remainingDataInTemporaryBuffer = 0;
            if (temporaryBuffer != null) {
                remainingDataInTemporaryBuffer = temporaryBuffer.remaining();
            }
            // Do we still need to complete writing the data from the temporary
            // buffer in order to complete an element in the buffers?
            if (remainingDataInTemporaryBuffer != 0) {
                // remainingInBuffer > remainingOffset, thus the cast is safe.
                int bytesToCopy = Math.min(remainingDataInTemporaryBuffer,
                        remainingInBuffer - (int) remainingOffset);
                copyBytes(temporaryBuffer, buffer, buffer.position()
                        + (int) remainingOffset, bytesToCopy);
                remainingOffset += bytesToCopy;
                remainingLength -= bytesToCopy;
                remainingUsableBytesInBuffer -= bytesToCopy;
                remainingDataInTemporaryBuffer -= bytesToCopy;
                // We never clear the buffer here. If we cleared the buffer, it
                // would look like there is data available in the next
                // iteration.
            }
            // If the remaining length is zero, we are done.
            if (remainingLength == 0) {
                break;
            }
            // If the current buffer has more remaining bytes than the element
            // size, we can write at least one element (possibly more).
            if (remainingUsableBytesInBuffer >= elementSize) {
                // We can write as many elements as still needed or as many
                // elements as there is space for in the current buffer -
                // whichever is less.
                int numberOfElementsToWrite = (int) Math.min(remainingLength,
                        remainingUsableBytesInBuffer) / elementSize;
                int numberOfBytesToWrite = numberOfElementsToWrite
                        * elementSize;
                ByteBuffer positionedBuffer = buffer.duplicate();
                // remainingInBuffer > remainingOffset, thus the cast is safe.
                positionedBuffer.position(positionedBuffer.position()
                        + (int) remainingOffset);
                valueWriter.writeValues(source, sourceOffset, positionedBuffer,
                        numberOfBytesToWrite);
                sourceOffset += numberOfElementsToWrite;
                remainingOffset += numberOfBytesToWrite;
                remainingLength -= numberOfBytesToWrite;
                remainingUsableBytesInBuffer -= numberOfBytesToWrite;
            }
            // If the remaining length is zero, we are done.
            if (remainingLength == 0) {
                break;
            }
            // If there is some space remaining in the buffer (not enough for a
            // complete element - otherwise we would have written it in the last
            // block - we copy an element into the temporary buffer and write a
            // part of it.
            if (remainingUsableBytesInBuffer != 0) {
                // We know that the temporary buffer must be empty, because
                // otherwise the available space would have been used to
                // write it. Therefore, we just use an assertion here.
                assert (remainingDataInTemporaryBuffer == 0);
                // The temporary buffer might not be initialized yet, so we have
                // to check.
                if (temporaryBuffer == null) {
                    // We use the byte order of the destination buffers. This is
                    // important because otherwise we might see unexpected
                    // results when writing multi-byte data-types.
                    temporaryBuffer = ByteBuffer.allocate(elementSize).order(
                            buffer.order());
                } else {
                    // If we already have a temporary buffer, we have to clear
                    // it.
                    temporaryBuffer.clear();
                }
                // We copy one element from the source array into the buffer.
                valueWriter.writeValues(source, sourceOffset, temporaryBuffer,
                        elementSize);
                sourceOffset += 1;
                // Prepare the temporary buffer for reading.
                temporaryBuffer.position(elementSize);
                temporaryBuffer.flip();
                remainingDataInTemporaryBuffer = elementSize;
                // Now we copy as much data into the target buffer as we can.
                // remainingInBuffer > remainingOffset, thus the cast is safe.
                copyBytes(temporaryBuffer, buffer, buffer.position()
                        + (int) remainingOffset, remainingUsableBytesInBuffer);
                // We do not have to set these variables here, because they are
                // either overwritten or discarded at the end of the iteration.
                // However, we still set the correct values here so that we have
                // a consistent state after this block, in case some additional
                // code is added at a later point in time.
                remainingOffset += remainingUsableBytesInBuffer;
                remainingLength -= remainingUsableBytesInBuffer;
                remainingDataInTemporaryBuffer -= remainingUsableBytesInBuffer;
                remainingUsableBytesInBuffer = 0;
            }
            // If the remaining length is zero, we are done.
            if (remainingLength == 0) {
                break;
            }
            // We have consumed all data from the current buffer but need the
            // all available data from the next buffer, thus the offset has to
            // be set to zero.
            remainingOffset = 0;
        }

        // We reached the end of the loop. If there is still data to be written,
        // we ran out of space.
        if (remainingLength != 0) {
            throw new BufferOverflowException();
        }
    }

    private ByteBuffer bufferFromIteratorOrNewBuffer(
            ListIterator<ByteBuffer> iterator, long requiredSpaceInBytes) {
        while (iterator.hasNext()) {
            ByteBuffer buffer = iterator.next();
            if (buffer != null) {
                return buffer;
            }
        }
        ByteBuffer buffer = requestAdditionalBuffer((requiredSpaceInBytes > Integer.MAX_VALUE) ? Integer.MAX_VALUE
                : (int) requiredSpaceInBytes);
        if (buffer != null) {
            if (buffer.position() != 0) {
                // The buffer's position should always be null. Otherwise, there
                // is already data in the buffer. We could throw an exception
                // here, but this exception would end up at the caller of the
                // originating put operation, who will not expect it. Therefore,
                // we just do not use the buffer and return null instead. This
                // will also lead to an exception (a BufferOverflowException),
                // however this exception is to be expected by the caller.
                return null;
            }
            iterator.add(buffer);
        }
        return buffer;
    }

    private void copyBytes(ByteBuffer sourceBuffer,
            ByteBuffer destinationBuffer, int positionInDestinationBuffer,
            int numberOfBytes) {
        // We do parameter checks as assertions only. The parameters should
        // always be valid.
        assert (destinationBuffer.remaining() >= numberOfBytes);
        assert (sourceBuffer.remaining() >= numberOfBytes);
        // We have to duplicate the destination buffer, so that we do not change
        // the original's buffer position.
        ByteBuffer duplicatedDestinationBuffer = destinationBuffer.duplicate();
        duplicatedDestinationBuffer.position(positionInDestinationBuffer);
        // We also have to duplicate the source buffer, so that we can write the
        // specified number of bytes. If we used the original buffer, we might
        // try to write more data than there is space for in the destination
        // buffer, causing a BufferOverflowException.
        ByteBuffer duplicatedSourceBuffer = sourceBuffer.duplicate();
        int sourceBufferPosition = duplicatedSourceBuffer.position();
        duplicatedSourceBuffer.position(sourceBufferPosition + numberOfBytes);
        duplicatedSourceBuffer.flip();
        duplicatedSourceBuffer.position(sourceBufferPosition);
        duplicatedDestinationBuffer.put(duplicatedSourceBuffer);
        // Now we advance the position on the original source buffer to mark the
        // data as read.
        sourceBuffer.position(sourceBufferPosition + numberOfBytes);
    }

    /**
     * <p>
     * Called when data has been committed to the byte buffers backing this byte
     * sink. This method is called every time an atomic put operation finishes
     * successfully (or a simple put operation outside an atomic put finishes).
     * It is called after the data has been committed, so a call to
     * {@link #getWrittenData()} returns the data that has just been written.
     * </p>
     * 
     * <p>
     * Derived classes can override this method when they want to be notified
     * each time data is written. The base implementation is empty, so this
     * method does not have to be called by overriding implementations.
     * </p>
     */
    protected void dataWritten() {
        // Derived classes can override this method to act on new data being
        // written to this byte sink.
    }

    private void commit(long length) {
        if (length <= 0L) {
            return;
        }
        long remainingLength = length;
        while (remainingLength > 0L) {
            // It is safe to access the buffer-list without checking it, because
            // the write operations have been successful, thus there must be
            // sufficient space in the buffers.
            ByteBuffer nextBuffer = nextBuffers.peek();
            int remainingInNextBuffer = nextBuffer.remaining();
            if (remainingInNextBuffer > remainingLength) {
                nextBuffer.position(nextBuffer.position()
                        + (int) remainingLength);
                bytesWritten += remainingLength;
                // Now the first buffer in the list is "dirty" because we have
                // committed data to it, but it has not been moved to the list
                // of full buffers yet.
                nextBufferIsDirty = true;
                dataWritten();
                return;
            }
            nextBuffer.position(nextBuffer.position() + remainingInNextBuffer);
            fullBuffers.addLast(nextBuffers.pop());
            remainingLength -= remainingInNextBuffer;
            bytesWritten += remainingInNextBuffer;
            // Now that we have removed the next buffer from the list, we can
            // clear the dirty flag, because the next buffer is now clean. The
            // dirty flag will be set again, if we commit data to the new next
            // buffer.
            nextBufferIsDirty = false;
        }
        dataWritten();
    }

    @Override
    public <T> T atomicPut(AtomicPutOperation<? extends T> atomicPutOperation) {
        // If there is already an atomic put operation in progress, we push the
        // current write offset to the stack. We will need this offset if we
        // want to roll back the put operation we are starting. If there is no
        // atomic put operation in progress yet, we start one.
        if (atomicWriteInProgress) {
            atomicWriteOffsetStack.push(atomicWriteOffset);
        } else {
            // Setting the atomicWriteInProgress flag has two consequences: Puts
            // are not automatically committed and we will start a nested atomic
            // put operation if this method is called again.
            atomicWriteInProgress = true;
        }
        // Depending on whether the operation was successful, we either commit
        // the put operations or we roll them back.
        boolean successful = false;
        try {
            T result = atomicPutOperation.put();
            successful = true;
            return result;
        } finally {
            if (successful) {
                // If the put operations were successful and this is not a
                // nested atomic put operation, we commit all operations. If
                // this is a nested operation, we do not commit and just
                // continue with the new offset. This has the same effect as if
                // all operations in the nested operation would have been
                // performed by the parent operation.
                if (atomicWriteOffsetStack.isEmpty()) {
                    commit(atomicWriteOffset);
                    atomicWriteOffset = 0;
                    atomicWriteInProgress = false;
                } else {
                    atomicWriteOffsetStack.pop();
                }
            } else {
                // If the put operations were not successful, we have to roll
                // back. For non-nested atomic put operation, we just reset the
                // offset and mark the operation as finished. For a nested
                // operation, we reset the offset to the value it had before
                // starting the nested operation. This has the same effect as if
                // the nested operation never happened.
                if (atomicWriteOffsetStack.isEmpty()) {
                    atomicWriteOffset = 0;
                    atomicWriteInProgress = false;
                } else {
                    atomicWriteOffset = atomicWriteOffsetStack.pop();
                }
            }
        }
    }

    private void checkSourceArrayLength(int arrayLength, int offset, int length) {
        if (offset < 0) {
            throw new IllegalArgumentException("Offset must not be negative.");
        }
        if (length < 0) {
            throw new IllegalArgumentException("Length must not be negative.");
        }
        if ((long) offset + (long) length > (long) arrayLength) {
            throw new IndexOutOfBoundsException();
        }
    }

    @Override
    public ByteBufferByteSink putByteArray(byte[] source, int offset, int length) {
        checkSourceArrayLength(source.length, offset, length);
        writeValues(length, SIZE_OF_BYTE, source, offset, BYTE_WRITER);
        return this;
    }

    @Override
    public ByteSink putDoubleArray(double[] source, int offset, int length) {
        checkSourceArrayLength(source.length, offset, length);
        writeValues(length, SIZE_OF_DOUBLE, source, offset, DOUBLE_WRITER);
        return this;
    }

    @Override
    public ByteSink putFloatArray(float[] source, int offset, int length) {
        checkSourceArrayLength(source.length, offset, length);
        writeValues(length, SIZE_OF_FLOAT, source, offset, FLOAT_WRITER);
        return this;
    }

    @Override
    public ByteSink putIntArray(int[] source, int offset, int length) {
        checkSourceArrayLength(source.length, offset, length);
        writeValues(length, SIZE_OF_INT, source, offset, INT_WRITER);
        return this;
    }

    @Override
    public ByteSink putShortArray(short[] source, int offset, int length) {
        checkSourceArrayLength(source.length, offset, length);
        writeValues(length, SIZE_OF_SHORT, source, offset, SHORT_WRITER);
        return this;
    }

    @Override
    public ByteSink putLongArray(long[] source, int offset, int length) {
        checkSourceArrayLength(source.length, offset, length);
        writeValues(length, SIZE_OF_LONG, source, offset, LONG_WRITER);
        return this;
    }

    private static void addByteToStringBuilder(StringBuilder sb, byte b) {
        sb.append("0x");
        if (b < 16 && b >= 0) {
            sb.append('0');
        }
        sb.append(Integer.toHexString(b & 0xff));
    }

    /**
     * Returns a string representation of this byte sink. The string
     * representation contains the last 1024 bytes that have been written to
     * this byte sink (provided that they are still in the byte sink and the
     * corresponding buffers have not been removed yet) in a hexadecimal
     * representation.
     * 
     * @return string representation of this byte sink.
     */
    @Override
    public String toString() {
        // This method is only intended for debugging purposes. Therefore, we
        // choose a rather simple (and not necessarily efficient)
        // implementation.
        final StringBuilder sb = new StringBuilder();
        sb.append(this.getClass().getName());
        sb.append("[");
        // Typically, we are most interested in the bytes written last.
        // Therefore, we limit ourselves to the last 1024 bytes.
        // If an atomic write is in progress, we have to consider the data from
        // the nextBuffers first.
        int sizeRead = 0;
        int sizeToReadFromNextBuffers = 0;
        int sizeToReadFromFullBuffers = 1024;
        if (atomicWriteInProgress) {
            if (atomicWriteOffset > sizeToReadFromFullBuffers) {
                sizeToReadFromNextBuffers = sizeToReadFromFullBuffers;
                sizeToReadFromFullBuffers = 0;
            } else {
                // The cast to int is safe because we know that
                // atomicWriteOffset <= sizeToReadFromFullBuffers.
                sizeToReadFromNextBuffers = (int) atomicWriteOffset;
                sizeToReadFromFullBuffers -= atomicWriteOffset;
            }
        }
        if (sizeToReadFromFullBuffers > 0) {
            LinkedList<ByteBuffer> buffers = new LinkedList<ByteBuffer>();
            int sizeFound = 0;
            ByteBuffer buffer;
            // The data that has already been committed to the next buffer has
            // to be treated like data in full buffers, thus we add this buffer
            // to the list.
            buffer = nextBuffers.peek();
            if (buffer != null) {
                // We have to duplicate the buffer so that we do not modify it.
                buffer = buffer.duplicate();
                // We flip the buffer so that we can read the data that has been
                // committed.
                buffer.flip();
                buffers.push(buffer);
                sizeFound += buffer.remaining();
            }
            // We have to traverse the full buffers in reverse order, because we
            // have to consider more recent data first.
            for (ListIterator<ByteBuffer> bufferIterator = fullBuffers
                    .listIterator(fullBuffers.size()); bufferIterator
                    .hasPrevious();) {
                // If we found more bytes than we need, we are done. If we found
                // exactly the number of bytes that we want, we continue,
                // because we have to no whether there is more data.
                if (sizeFound > sizeToReadFromFullBuffers) {
                    break;
                }
                // Decrement the iterator.
                buffer = bufferIterator.previous();
                // We have to duplicate the buffer so that we do not modify it.
                buffer = buffer.duplicate();
                // We flip the buffer so that we can read the data that has been
                // committed.
                buffer.flip();
                buffers.push(buffer);
                sizeFound += buffer.remaining();
            }
            if (sizeFound > sizeToReadFromFullBuffers) {
                // We want to show in the output that there is more data in the
                // buffers.
                sb.append("..., ");
                // We skip the extra bytes. We can safely assume that the number
                // of bytes remaining in the first buffer is greater than or
                // equal to the number of extra bytes, because we stopped adding
                // buffers once we got more bytes than we wanted.
                buffer = buffers.peek();
                buffer.position(buffer.position()
                        + (sizeFound - sizeToReadFromFullBuffers));
            }
            // Now we can print all data from the buffers. This should be
            // sizeToReadFromFullBuffers bytes or less in total.
            for (ByteBuffer fullBuffer : buffers) {
                while (fullBuffer.hasRemaining()) {
                    addByteToStringBuilder(sb, fullBuffer.get());
                    sb.append(", ");
                    sizeRead++;
                }
            }
        }
        if (sizeToReadFromNextBuffers > 0) {
            int sizeReadFromNextBuffers = 0;
            for (ByteBuffer buffer : nextBuffers) {
                // If we have read enough data, we can exit the loop.
                if (sizeReadFromNextBuffers == sizeToReadFromNextBuffers) {
                    break;
                }
                // We have to duplicate the buffer so that we do not modify it.
                buffer = buffer.duplicate();
                while (buffer.hasRemaining()
                        && sizeReadFromNextBuffers < sizeToReadFromNextBuffers) {
                    addByteToStringBuilder(sb, buffer.get());
                    sb.append(", ");
                    sizeReadFromNextBuffers++;
                }
            }
            sizeRead += sizeReadFromNextBuffers;
        }
        if (sizeRead > 0) {
            // We addes at least one byte, thus we can remove the final comma
            // and space.
            sb.delete(sb.length() - 2, sb.length());
        }
        sb.append(']');
        return sb.toString();
    }

}
