/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.Inet4AddressUtil;

/**
 * Representation of a <code>CA_PROTO_RSRV_IS_UP</code> message. This message is
 * sent from a server to a repeater and than relayed to the clients in order to
 * signal that the server is alive. Clients can use this information in order to
 * look for channels that could not be connected yet.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_RSRV_IS_UP
 */
public final class ChannelAccessBeaconMessage extends ChannelAccessMessage {

    private Inet4Address serverAddress;
    private int serverPort;
    private ChannelAccessVersion version;
    private int sequenceNumber;

    /**
     * Constructor for creating a <code>CA_PROTO_RSRV_IS_UP</code> message. This
     * constructor should be used for messages that are sent from a server to a
     * repeater.
     * 
     * @param serverAddress
     *            the IP address of the server sending this message. If this is
     *            the wildcard address, the receiving repeater will replace it
     *            with the address it received the message from.
     * @param serverPort
     *            the TCP port number on which the server is listening. If this
     *            parameter is zero, the receiving client should use the
     *            standard TCP port number (<code>EPICS_CA_SERVER_PORT</code>)
     *            instead.
     * @param version
     *            Channel Access version supported by the server.
     * @param sequenceNumber
     *            sequence number. This number is incremented each time a server
     *            sends a beacon message. However, the same message might be
     *            sent over multiple interfaces, therefore the same beacon
     *            number might appear twice.
     */
    public ChannelAccessBeaconMessage(Inet4Address serverAddress,
            int serverPort, ChannelAccessVersion version, int sequenceNumber) {
        super(ChannelAccessCommand.CA_PROTO_RSRV_IS_UP);

        if (serverAddress == null) {
            throw new NullPointerException(
                    "The server address must not be null.");
        }
        this.serverAddress = serverAddress;
        this.serverPort = serverPort;
        this.version = version;
        this.sequenceNumber = sequenceNumber;
    }

    private ChannelAccessBeaconMessage(
            ChannelAccessMessageHeader messageHeader,
            Inet4Address serverAddress, int serverPort,
            ChannelAccessVersion version, int sequenceNumber) {
        super(ChannelAccessCommand.CA_PROTO_RSRV_IS_UP);

        this.serverAddress = serverAddress;
        this.serverPort = serverPort;
        this.version = version;
        this.sequenceNumber = sequenceNumber;
    }

    /**
     * Returns the IP address of the server that sent the message.
     * 
     * @return IP address of the server that sent the message.
     */
    public Inet4Address getServerAddress() {
        return serverAddress;
    }

    /**
     * Returns TCP port on which the server listens for incoming connections. If
     * the port number is zero, the server listens on the standard port.
     * 
     * @return TCP port the server is listening on or zero if it listens on the
     *         standard port.
     */
    public int getServerPort() {
        return serverPort;
    }

    /**
     * Returns the most recent Channel Access version supported by the server.
     * 
     * @return Channel Access protocol version supported by the server.
     */
    public ChannelAccessVersion getVersion() {
        return version;
    }

    /**
     * Returns the sequence number. This number is incremented each time a
     * server sends a beacon message. However, the same message might be sent
     * over multiple interfaces, therefore the same beacon number might appear
     * in multiple messages.
     * 
     * @return sequence number.
     */
    public int getSequenceNumber() {
        return sequenceNumber;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, this.version.getMinorVersionAsShort(),
                serverPort, sequenceNumber,
                Inet4AddressUtil.addressToInt(serverAddress.getAddress()),
                version, maxPayloadSize);
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessBeaconMessage other = (ChannelAccessBeaconMessage) obj;
        return new EqualsBuilder()
                .append(this.serverAddress, other.serverAddress)
                .append(this.serverPort, other.serverPort)
                .append(this.version, other.version)
                .append(this.sequenceNumber, other.sequenceNumber).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(serverAddress).append(serverPort).append(version)
                .append(sequenceNumber).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("serverAddress=");
        sb.append(serverAddress != null ? serverAddress.getHostAddress()
                : "null");
        sb.append(", serverPort=");
        sb.append(serverPort);
        sb.append(", version=");
        sb.append(version.toString());
        sb.append(", sequenceNumber=");
        sb.append(sequenceNumber);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param originAddress
     *            the address from which the message originated. This address is
     *            used to replace the server address in this message if the
     *            server address is the wildcard address.
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete message (including he payload if
     *             <code>headerOnly</code> is <code>false</code>).
     */
    protected static ChannelAccessBeaconMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            Inet4Address originAddress) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_RSRV_IS_UP
                .getCommandNumber());
        try {
            byteSource.skip(messageHeader.getPayloadSize());
        } catch (BufferUnderflowException e) {
            // We ignore a buffer underflow. The repeater protocol uses a single
            // message only per packet anyway, therefore we should always be
            // aligned for the next message, even if the payload size was wrong.
        }
        Inet4Address serverAddress;
        try {
            InetAddress tempAddress = InetAddress.getByAddress(Inet4AddressUtil
                    .addressToBytes(messageHeader.getContextSpecific()));
            if (!(tempAddress instanceof Inet4Address)) {
                throw new RuntimeException(
                        "Did not get IPv4 address where it was expected. Got address of type "
                                + tempAddress.getClass().getName()
                                + " instead.");
            }
            serverAddress = (Inet4Address) tempAddress;
        } catch (UnknownHostException e) {
            throw new RuntimeException(
                    "Cannot get inet address instance for address of four bytes.",
                    e);
        }
        if (serverAddress.isAnyLocalAddress() && originAddress != null) {
            serverAddress = originAddress;
        }

        return new ChannelAccessBeaconMessage(messageHeader, serverAddress,
                messageHeader.getCount(),
                ChannelAccessVersion.forMinorVersion(messageHeader
                        .getDataType()), messageHeader.getCID());
    }

}
