/*
 * Copyright 2014-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.net.Inet4Address;
import java.nio.BufferOverflowException;
import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import com.aquenos.epics.jackie.common.exception.BufferUnderflowWithSizeException;
import com.aquenos.epics.jackie.common.exception.MalformedMessageException;
import com.aquenos.epics.jackie.common.exception.ReceivedMessageTooLargeException;
import com.aquenos.epics.jackie.common.exception.SerializedChannelNameTooLargeException;
import com.aquenos.epics.jackie.common.exception.SerializedPayloadTooLargeException;
import com.aquenos.epics.jackie.common.exception.ShortPayloadException;
import com.aquenos.epics.jackie.common.exception.UnsupportedInThisVersionException;
import com.aquenos.epics.jackie.common.exception.UnsupportedMessageTypeException;
import com.aquenos.epics.jackie.common.exception.UnsupportedProtocolVersionException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSink.AtomicPutOperation;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.io.ByteSource.AtomicGetOperation;

/**
 * <p>
 * Encoder and decoder for Channel Access messages.
 * </p>
 * 
 * <p>
 * Messages can be encoded (serialized) using one of the <code>encode</code>
 * methods. They can be decoded (deserialized) using one of the
 * <code>decode</code> methods. These methods are safe for concurrent use by
 * multiple threads.
 * </p>
 * 
 * <p>
 * Problems that occur while encoding or decoding a message a signaled by
 * exceptions. When an exception is encoutered while trying to decode a message,
 * the corresponding message may be discarded from the byte source using the
 * {@link #discardNextMessage(ByteSource)} or
 * {@link #discardNextMessageCompletely(ByteSource)} method. In general, the
 * {@link #discardNextMessage(ByteSource)} method is preferrable because it will
 * work, even if the message is too large to fit in the buffer backing the byte
 * source. However, in contrast to the
 * {@link #discardNextMessageCompletely(ByteSource)} method, the calling code
 * has to perform additional work in order to ensure that the message is really
 * discarded completely.
 * </p>
 * 
 * <p>
 * The character set that is used for encoding and decoding strings in messages
 * can be configured with the {@link #setCharset(Charset)} method. However,
 * changing the configuration is not thread safe, therefore the calling code
 * must ensure that no other thread uses the message codec at the same time. By
 * default, a newly created codec uses the UTF-8 encoding or, if this encoding
 * is not available on the platform, the platform's default encoding.
 * </p>
 * 
 * <p>
 * The maximum allowed payload size can be configured separately for received
 * messages (that are decoded) and sent messages (that are encoded) using the
 * {@link #setMaxPayloadReceiveSize(int)} and
 * {@link #setMaxPayloadSendSize(int)} methods. However, changing the
 * configuration is not thread safe, therefore the calling code must ensure that
 * no other thread uses the message codec at the same time. By default, a
 * maximum payload size of {@value #MIN_MAX_PAYLOAD_SIZE} bytes is used. This is
 * also the smallest maximum payload size that can be configured. The Channel
 * Access C library (as of EPICS 3.15.0) uses the same default setting. The
 * upper limit for the maximum payload size is {@value #MAX_MAX_PAYLOAD_SIZE}
 * bytes. This limit comes from the fact that this library used 32-bit signed
 * integers for dealing with message sizes. The Channel Access protocol uses
 * 32-bit unsigned integers and thus would allow a greater limit, however no
 * reasonable message sent via Channel Access should exceed limit of this
 * implementation (which is roughly equivalent to a payload size of 2 GB)
 * anyway.
 * </p>
 * 
 * <p>
 * While this class's encode and decode methods are thread-safe, the methods for
 * changing the configuration are not thread-safe (not even with respect to the
 * encode and decode methods). Therefore, a thread has to ensure that it has
 * exclusive access to a message codec before changing its configuration.
 * Typically, the configuration is only changed once after creating a message
 * codec and never touched later.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class ChannelAccessMessageCodec {

    /**
     * Dummy exception. This exception is only thrown to abort an atomic get
     * operation. It does not indicate an error but actually indicates success.
     * 
     * @author Sebastian Marsching
     */
    private static class PayloadAvailableException extends RuntimeException {
        private static final long serialVersionUID = 5321542267146195576L;
    }

    /**
     * Smallest valid value for the maximum payload-size option.
     */
    public final static int MIN_MAX_PAYLOAD_SIZE = 0x4000;

    /**
     * Largest valid value for the maximum payload-size option.
     */
    public final static int MAX_MAX_PAYLOAD_SIZE = (Integer.MAX_VALUE - Integer.MAX_VALUE
            % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT)
            - ChannelAccessConstants.LARGE_HEADER_SIZE;

    private int maxPayloadReceiveSize = MIN_MAX_PAYLOAD_SIZE;
    private int maxPayloadSendSize = MIN_MAX_PAYLOAD_SIZE;
    private Charset charset;

    /**
     * Creates a message codec that uses the UTF-8 character set or, if this
     * character set is not available on the platform, the platform's default
     * character set.
     * 
     * The maximum payload send and receive sizes are initialized to their
     * smallest possible values ({@value #MIN_MAX_PAYLOAD_SIZE} bytes).
     */
    public ChannelAccessMessageCodec() {
        // It is a very unlikely that the UTF-8 charset is not supported.
        // However, if it is not supported, we still want to be able to create
        // instances of this class. Therefore we have a fall-back to the
        // platform's default charset.
        if (Charset.isSupported("UTF-8")) {
            charset = Charset.forName("UTF-8");
        } else {
            charset = Charset.defaultCharset();
        }
    }

    /**
     * Returns the maximum payload size for received messages. If a received
     * message's payload exceeds this size, the deserialization routine will
     * throw a {@link ReceivedMessageTooLargeException}. The default value is
     * {@value #MIN_MAX_PAYLOAD_SIZE}.
     * 
     * @return maximum payload size for received Channel Access messages (in
     *         bytes). This number is always a multiple of
     *         {@link ChannelAccessConstants#MESSAGE_SIZE_ALIGNMENT}.
     */
    public int getMaxPayloadReceiveSize() {
        return maxPayloadReceiveSize;
    }

    /**
     * <p>
     * Sets the maximum payload size for received messages. If a received
     * message's payload exceeds this size, the deserialization routine will
     * throw a {@link ReceivedMessageTooLargeException}. The default value is
     * {@value #MIN_MAX_PAYLOAD_SIZE} which is also the smallest allowed value.
     * </p>
     * <p>
     * <b>Caution:</b> This method is not thread-safe. It should be called at
     * initialization time, before the message codec is used. If the value is
     * supposed to be changed later, it has to be ensured that no other thread
     * is using this message codec concurrently.
     * </p>
     * 
     * @param maxPayloadReceiveSize
     *            new maxium payload size for received Channel Access messages
     *            (in bytes). Must be at least {@value #MIN_MAX_PAYLOAD_SIZE}.
     *            Must not be greater than {@value #MAX_MAX_PAYLOAD_SIZE}. If
     *            this number is not a multiple of
     *            {@link ChannelAccessConstants#MESSAGE_SIZE_ALIGNMENT}, the
     *            next smaller number that is a multiple of
     *            <code>MESSAGE_SIZE_ALIGNMENT</code> is used.
     * @throws IllegalArgumentException
     *             if <code>maxPayloadReceiveSize</code> is less than
     *             {@value #MIN_MAX_PAYLOAD_SIZE} or greater than
     *             {@value #MAX_MAX_PAYLOAD_SIZE}.
     */
    public void setMaxPayloadReceiveSize(int maxPayloadReceiveSize) {
        if (maxPayloadReceiveSize < MIN_MAX_PAYLOAD_SIZE) {
            throw new IllegalArgumentException(
                    "Maximum payload-size must be at least "
                            + MIN_MAX_PAYLOAD_SIZE + ".");
        }
        if (maxPayloadReceiveSize > MAX_MAX_PAYLOAD_SIZE) {
            throw new IllegalArgumentException(
                    "Maximum payload-size must not be greater than "
                            + MAX_MAX_PAYLOAD_SIZE);
        }
        this.maxPayloadReceiveSize = maxPayloadReceiveSize
                - maxPayloadReceiveSize
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT;
    }

    /**
     * Returns the maximum payload size for sent messages. If a message's
     * payload exceeds this size, the serialization routine will throw a
     * {@link SerializedPayloadTooLargeException}. The default value is
     * {@value #MIN_MAX_PAYLOAD_SIZE}.
     * 
     * @return maximum payload size for sent Channel Access messages (in bytes).
     *         This number is always a multiple of
     *         {@link ChannelAccessConstants#MESSAGE_SIZE_ALIGNMENT}.
     */
    public int getMaxPayloadSendSize() {
        return maxPayloadSendSize;
    }

    /**
     * <p>
     * Sets the maximum payload size for sent messages. If a message's payload
     * exceeds this size, the serialization routine will throw a
     * {@link SerializedPayloadTooLargeException}. The default value is
     * {@value #MIN_MAX_PAYLOAD_SIZE} which is also the smallest allowed value.
     * </p>
     * <p>
     * <b>Caution:</b> This method is not thread-safe. It should be called at
     * initialization time, before the message codec is used. If the value is
     * supposed to be changed later, it has to be ensured that no other thread
     * is using this message codec concurrently.
     * </p>
     * 
     * @param maxPayloadSendSize
     *            new maxium payload size for sent Channel Access messages (in
     *            bytes). Must be at least {@value #MIN_MAX_PAYLOAD_SIZE}. Must
     *            not be greater than {@value #MAX_MAX_PAYLOAD_SIZE}. If this
     *            number is not a multiple of
     *            {@link ChannelAccessConstants#MESSAGE_SIZE_ALIGNMENT}, the
     *            next smaller number that is a multiple of
     *            <code>MESSAGE_SIZE_ALIGNMENT</code> is used.
     * @throws IllegalArgumentException
     *             if <code>maxPayloadSendSize</code> is less than
     *             {@value #MIN_MAX_PAYLOAD_SIZE} or greater than
     *             {@value #MAX_MAX_PAYLOAD_SIZE}.
     */
    public void setMaxPayloadSendSize(int maxPayloadSendSize) {
        if (maxPayloadSendSize < MIN_MAX_PAYLOAD_SIZE) {
            throw new IllegalArgumentException(
                    "Maximum payload-size must be at least "
                            + MIN_MAX_PAYLOAD_SIZE + ".");
        }
        this.maxPayloadSendSize = maxPayloadSendSize - maxPayloadSendSize
                % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT;
    }

    /**
     * Returns the character set used by this message codec. The character set
     * determines how strings are converted to bytes and the other-way round.
     * This affects the over-the-wire format of channel names, error strings and
     * string values. The default is the UTF-8 charset. If this charset is not
     * available, the default is the platform's default charset.
     * 
     * @return character-set used for encoding and decoding strings.
     */
    public Charset getCharset() {
        return charset;
    }

    /**
     * <p>
     * Sets the character set used by this message codec. The character set
     * determines how strings are converted to bytes and the other-way round.
     * This affects the over-the-wire format of channel names, error strings and
     * string values. The default is the UTF-8 charset. If this charset is not
     * available, the default is the platform's default charset.
     * </p>
     * <p>
     * <b>Caution:</b> This method is not thread-safe. It should be called at
     * initialization time, before the message codec is used. If the value is
     * supposed to be changed later, it has to be ensured that no other thread
     * is using this message codec concurrently.
     * </p>
     * 
     * @param charset
     *            new character-set to use for encoding and decoding strings.
     */
    public void setCharset(Charset charset) {
        // If the encoding is null, we rather throw an exception here than the
        // first time we use the encoding.
        if (charset == null) {
            throw new NullPointerException("Charset must not be null.");
        }
        this.charset = charset;
    }

    private void encodeMessageInternal(final ByteSink byteSink,
            final ChannelAccessMessage message,
            final ChannelAccessVersion version) {
        // We run the serialization within a transaction. This way, either the
        // complete message or nothing is written.
        byteSink.atomicPut(new AtomicPutOperation<Void>() {

            @Override
            public Void put() {
                message.serialize(byteSink, version, maxPayloadSendSize,
                        charset);
                return null;
            }
        });
    }

    /**
     * Encodes a message that shall be sent from a client to a server via UDP.
     * Only messages of type {@link ChannelAccessVersionUDPMessage} and
     * {@link ChannelAccessSearchClientMessage} can be sent via this protocol.
     * 
     * @param byteSink
     *            byte sink the encoded message is written to. The message is
     *            written in an atomic operation, so it is either written
     *            completely or nothing is written at all.
     * @param message
     *            message to be encoded.
     * @param version
     *            protocol version that shall be used. This implementation only
     *            supports protocol versions equal to or greater than
     *            {@link ChannelAccessVersion#V4_4}.
     * @throws BufferOverflowException
     *             if this exception is thrown by the byte sink and thus the
     *             message cannot be written completely.
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent from a
     *             client to a server via UDP.
     * @throws SerializedChannelNameTooLargeException
     *             if a channel name has to be serialized and the serialized
     *             form of the channel name is too long.
     * @throws SerializedPayloadTooLargeException
     *             if the message cannot be sent because the payload would be
     *             larger than the configured maximum payload send size or the
     *             specified protocol version is older than Channel Access
     *             version 4.9 and the payload would be larger than 65528 bytes
     *             or the count exceeds 65535 elements.
     * @throws UnsupportedInThisVersionException
     *             if this message is not support in the protocol version
     *             specified by <code>version</code>.
     * @throws UnsupportedOperationException
     *             if this message has been constructed by deserializing it from
     *             headers only and thus it is incomplete.
     * @throws UnsupportedProtocolVersionException
     *             if the specified <code>version</code> is older than the
     *             oldest supported protocol version for this message. In
     *             general, all Channel Access versions since version 4.4 should
     *             be supported.
     */
    public void encodeMessageToUDPServer(ByteSink byteSink,
            ChannelAccessMessage message, ChannelAccessVersion version) {
        if (message instanceof ChannelAccessVersionUDPMessage
                || message instanceof ChannelAccessSearchClientMessage) {
            encodeMessageInternal(byteSink, message, version);
        } else {
            // If we got an unsupported message type for this type of transport,
            // this must be a bug in the implementation, therefore throwing an
            // IllegalArgumentException seems appropriate.
            throw new IllegalArgumentException("Cannot send message of type "
                    + message.getClass().getName() + " to UDP server.");
        }
    }

    /**
     * Encodes a message that shall be sent from a server to a client via UDP.
     * Only messages of type {@link ChannelAccessVersionUDPMessage} and
     * {@link ChannelAccessSearchUDPServerMessage},
     * {@link ChannelAccessEchoMessage}, {@link ChannelAccessErrorMessage}, and
     * {@link ChannelAccessNotFoundMessage} can be sent via this protocol.
     * 
     * @param byteSink
     *            byte sink the encoded message is written to. The message is
     *            written in an atomic operation, so it is either written
     *            completely or nothing is written at all.
     * @param message
     *            message to be encoded.
     * @param version
     *            protocol version that shall be used. This implementation only
     *            supports protocol versions equal to or greater than
     *            {@link ChannelAccessVersion#V4_4}.
     * @throws BufferOverflowException
     *             if this exception is thrown by the byte sink and thus the
     *             message cannot be written completely.
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent from a
     *             server to a client via UDP.
     * @throws SerializedPayloadTooLargeException
     *             if the message cannot be sent because the payload would be
     *             larger than the configured maximum payload send size or the
     *             specified protocol version is older than Channel Access
     *             version 4.9 and the payload would be larger than 65528 bytes
     *             or the count exceeds 65535 elements.
     * @throws UnsupportedInThisVersionException
     *             if this message is not support in the protocol version
     *             specified by <code>version</code>.
     * @throws UnsupportedOperationException
     *             if this message has been constructed by deserializing it from
     *             headers only and thus it is incomplete.
     * @throws UnsupportedProtocolVersionException
     *             if the specified <code>version</code> is older than the
     *             oldest supported protocol version for this message. In
     *             general, all Channel Access versions since version 4.4 should
     *             be supported.
     */
    public void encodeMessageToUDPClient(ByteSink byteSink,
            ChannelAccessMessage message, ChannelAccessVersion version) {
        if (message instanceof ChannelAccessVersionUDPMessage
                || message instanceof ChannelAccessSearchUDPServerMessage
                || message instanceof ChannelAccessEchoMessage
                || message instanceof ChannelAccessErrorMessage
                || message instanceof ChannelAccessNotFoundMessage) {
            encodeMessageInternal(byteSink, message, version);
        } else {
            // If we got an unsupported message type for this type of transport,
            // this must be a bug in the implementation, therefore throwing an
            // IllegalArgumentException seems appropriate.
            throw new IllegalArgumentException("Cannot send message of type "
                    + message.getClass().getName() + " to UDP client.");
        }
    }

    /**
     * Encodes a message that shall be sent from a server or a client to a
     * repeater via UDP. Only messages of type
     * {@link ChannelAccessBeaconMessage} can be sent via this protocol.
     * 
     * @param byteSink
     *            byte sink the encoded message is written to. The message is
     *            written in an atomic operation, so it is either written
     *            completely or nothing is written at all.
     * @param message
     *            message to be encoded.
     * @param version
     *            protocol version that shall be used. This implementation only
     *            supports protocol versions equal to or greater than
     *            {@link ChannelAccessVersion#V4_4}.
     * @throws BufferOverflowException
     *             if this exception is thrown by the byte sink and thus the
     *             message cannot be written completely.
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent to a
     *             repeater.
     */
    public void encodeMessageToRepeater(ByteSink byteSink,
            ChannelAccessMessage message, ChannelAccessVersion version) {
        // We could also support ChannelAccessRepeaterRegisterMessage here, but
        // this would just result in an UnsupportedOperationException when we
        // call the serialize method.
        if (message instanceof ChannelAccessBeaconMessage) {
            encodeMessageInternal(byteSink, message, version);
        } else {
            // If we got an unsupported message type for this type of transport,
            // this must be a bug in the implementation, therefore throwing an
            // IllegalArgumentException seems appropriate.
            throw new IllegalArgumentException("Cannot send message of type "
                    + message.getClass().getName() + " to repeater.");
        }
    }

    /**
     * Encodes a message that shall be sent from a repeater to a client via UDP.
     * Only messages of type {@link ChannelAccessVersionUDPMessage},
     * {@link ChannelAccessRepeaterConfirmMessage}, and
     * {@link ChannelAccessBeaconMessage} can be sent via this protocol.
     * 
     * @param byteSink
     *            byte sink the encoded message is written to. The message is
     *            written in an atomic operation, so it is either written
     *            completely or nothing is written at all.
     * @param message
     *            message to be encoded.
     * @param version
     *            protocol version that shall be used. This implementation only
     *            supports protocol versions equal to or greater than
     *            {@link ChannelAccessVersion#V4_4}.
     * @throws BufferOverflowException
     *             if this exception is thrown by the byte sink and thus the
     *             message cannot be written completely.
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent from a
     *             repeater to a client via UDP.
     */
    public void encodeMessageFromRepeater(ByteSink byteSink,
            ChannelAccessMessage message, ChannelAccessVersion version) {
        if (message instanceof ChannelAccessVersionUDPMessage
                || message instanceof ChannelAccessRepeaterConfirmMessage
                || message instanceof ChannelAccessBeaconMessage) {
            encodeMessageInternal(byteSink, message, version);
        } else {
            // If we got an unsupported message type for this type of transport,
            // this must be a bug in the implementation, therefore throwing an
            // IllegalArgumentException seems appropriate.
            throw new IllegalArgumentException("Cannot send message of type "
                    + message.getClass().getName() + " from repeater.");
        }
    }

    /**
     * Encodes a message that shall be sent from a client to a server via TCP
     * Only messages of type {@link ChannelAccessVersionTCPClientMessage},
     * {@link ChannelAccessSubscriptionClientMessage},
     * {@link ChannelAccessCancelSubscriptionClientMessage},
     * {@link ChannelAccessReadClientMessage}, {@link ChannelAccessWriteMessage}
     * , {@link ChannelAccessEventsOffMessage},
     * {@link ChannelAccessEventsOnMessage},
     * {@link ChannelAccessReadSyncMessage},
     * {@link ChannelAccessDisconnectChannelMessage},
     * {@link ChannelAccessReadNotifyClientMessage},
     * {@link ChannelAccessConnectChannelClientMessage},
     * {@link ChannelAccessWriteNotifyClientMessage},
     * {@link ChannelAccessUserNameMessage},
     * {@link ChannelAccessHostNameMessage}, {@link ChannelAccessEchoMessage},
     * and {@link ChannelAccessSearchClientMessage} can be sent via this
     * protocol.
     * 
     * @param byteSink
     *            byte sink the encoded message is written to. The message is
     *            written in an atomic operation, so it is either written
     *            completely or nothing is written at all.
     * @param message
     *            message to be encoded.
     * @param version
     *            protocol version that shall be used. This implementation only
     *            supports protocol versions equal to or greater than
     *            {@link ChannelAccessVersion#V4_4}.
     * @throws BufferOverflowException
     *             if this exception is thrown by the byte sink and thus the
     *             message cannot be written completely.
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent from a
     *             client to a server via TCP.
     * @throws SerializedPayloadTooLargeException
     *             if the message cannot be sent because the payload would be
     *             larger than the configured maximum payload send size or the
     *             specified protocol version is older than Channel Access
     *             version 4.9 and the payload would be larger than 65528 bytes
     *             or the count exceeds 65535 elements.
     * @throws UnsupportedInThisVersionException
     *             if this message is not support in the protocol version
     *             specified by <code>version</code>.
     * @throws UnsupportedOperationException
     *             if this message has been constructed by deserializing it from
     *             headers only and thus it is incomplete.
     * @throws UnsupportedProtocolVersionException
     *             if the specified <code>version</code> is older than the
     *             oldest supported protocol version for this message. In
     *             general, all Channel Access versions since version 4.4 should
     *             be supported.
     */
    public void encodeMessageToTCPServer(ByteSink byteSink,
            ChannelAccessMessage message, ChannelAccessVersion version) {
        // We could also support ChannelAccessReadBuildMessage here, but this
        // would just result in an UnsupportedOperationException when we call
        // the serialize method.
        if (message instanceof ChannelAccessVersionTCPClientMessage
                || message instanceof ChannelAccessSubscriptionClientMessage
                || message instanceof ChannelAccessCancelSubscriptionClientMessage
                || message instanceof ChannelAccessReadClientMessage
                || message instanceof ChannelAccessWriteMessage
                || message instanceof ChannelAccessEventsOffMessage
                || message instanceof ChannelAccessEventsOnMessage
                || message instanceof ChannelAccessReadSyncMessage
                || message instanceof ChannelAccessDisconnectChannelMessage
                || message instanceof ChannelAccessReadNotifyClientMessage
                || message instanceof ChannelAccessConnectChannelClientMessage
                || message instanceof ChannelAccessWriteNotifyClientMessage
                || message instanceof ChannelAccessUserNameMessage
                || message instanceof ChannelAccessHostNameMessage
                || message instanceof ChannelAccessEchoMessage
                || message instanceof ChannelAccessSearchClientMessage) {
            encodeMessageInternal(byteSink, message, version);
        } else {
            // If we got an unsupported message type for this type of transport,
            // this must be a bug in the implementation, therefore throwing an
            // IllegalArgumentException seems appropriate.
            throw new IllegalArgumentException("Cannot send message of type "
                    + message.getClass().getName() + " to TCP server.");
        }
    }

    /**
     * Encodes a message that shall be sent from a server to a client via TCP
     * Only messages of type {@link ChannelAccessVersionTCPServerMessage},
     * {@link ChannelAccessSubscriptionServerMessage},
     * {@link ChannelAccessCancelSubscriptionServerMessage},
     * {@link ChannelAccessReadServerMessage},
     * {@link ChannelAccessReadSyncMessage} , {@link ChannelAccessErrorMessage},
     * {@link ChannelAccessDisconnectChannelMessage},
     * {@link ChannelAccessReadNotifyServerMessage},
     * {@link ChannelAccessConnectChannelServerMessage},
     * {@link ChannelAccessWriteNotifyServerMessage},
     * {@link ChannelAccessAccessRightsMessage},
     * {@link ChannelAccessEchoMessage},
     * {@link ChannelAccessConnectChannelFailedMessage},
     * {@link ChannelAccessChannelDisconnectedByServerMessage},
     * {@link ChannelAccessSearchTCPServerMessage}, and
     * {@link ChannelAccessNotFoundMessage} can be sent via this protocol.
     * 
     * @param byteSink
     *            byte sink the encoded message is written to. The message is
     *            written in an atomic operation, so it is either written
     *            completely or nothing is written at all.
     * @param message
     *            message to be encoded.
     * @param version
     *            protocol version that shall be used. This implementation only
     *            supports protocol versions equal to or greater than
     *            {@link ChannelAccessVersion#V4_4}.
     * @throws BufferOverflowException
     *             if this exception is thrown by the byte sink and thus the
     *             message cannot be written completely.
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent from a
     *             server to a client via TCP.
     * @throws SerializedPayloadTooLargeException
     *             if the message cannot be sent because the payload would be
     *             larger than the configured maximum payload send size or the
     *             specified protocol version is older than Channel Access
     *             version 4.9 and the payload would be larger than 65528 bytes
     *             or the count exceeds 65535 elements.
     * @throws UnsupportedInThisVersionException
     *             if this message is not support in the protocol version
     *             specified by <code>version</code>.
     * @throws UnsupportedOperationException
     *             if this message has been constructed by deserializing it from
     *             headers only and thus it is incomplete.
     * @throws UnsupportedProtocolVersionException
     *             if the specified <code>version</code> is older than the
     *             oldest supported protocol version for this message. In
     *             general, all Channel Access versions since version 4.4 should
     *             be supported.
     */
    public void encodeMessageToTCPClient(ByteSink byteSink,
            ChannelAccessMessage message, ChannelAccessVersion version) {
        // We could also support ChannelAccessReadBuildMessage here, but this
        // would just result in an UnsupportedOperationException when we call
        // the serialize method.
        if (message instanceof ChannelAccessVersionTCPServerMessage
                || message instanceof ChannelAccessSubscriptionServerMessage
                || message instanceof ChannelAccessCancelSubscriptionServerMessage
                || message instanceof ChannelAccessReadServerMessage
                || message instanceof ChannelAccessReadSyncMessage
                || message instanceof ChannelAccessErrorMessage
                || message instanceof ChannelAccessDisconnectChannelMessage
                || message instanceof ChannelAccessReadNotifyServerMessage
                || message instanceof ChannelAccessConnectChannelServerMessage
                || message instanceof ChannelAccessWriteNotifyServerMessage
                || message instanceof ChannelAccessAccessRightsMessage
                || message instanceof ChannelAccessEchoMessage
                || message instanceof ChannelAccessConnectChannelFailedMessage
                || message instanceof ChannelAccessChannelDisconnectedByServerMessage
                || message instanceof ChannelAccessSearchTCPServerMessage
                || message instanceof ChannelAccessNotFoundMessage) {
            encodeMessageInternal(byteSink, message, version);
        } else {
            // If we got an unsupported message type for this type of transport,
            // this must be a bug in the implementation, therefore throwing an
            // IllegalArgumentException seems appropriate.
            throw new IllegalArgumentException("Cannot send message of type "
                    + message.getClass().getName() + " to TCP client.");
        }
    }

    private static ChannelAccessMessageHeader decodeMessageHeader(
            ByteSource byteSource) {
        short command, payloadSize, dataType, count;
        int cid, contextSpecific;
        try {
            command = byteSource.getShort();
            payloadSize = byteSource.getShort();
            dataType = byteSource.getShort();
            count = byteSource.getShort();
            cid = byteSource.getInt();
            contextSpecific = byteSource.getInt();
        } catch (BufferUnderflowException e) {
            throw new BufferUnderflowWithSizeException(
                    ChannelAccessConstants.STANDARD_HEADER_SIZE);
        }
        int largePayloadSize;
        int largeCount;
        boolean largeHeader;
        // A payload size of 0xffff signals a large header.
        if (payloadSize == (short) 0xffff) {
            largeHeader = true;
            try {
                largePayloadSize = byteSource.getInt();
                largeCount = byteSource.getInt();
            } catch (BufferUnderflowException e) {
                throw new BufferUnderflowWithSizeException(
                        ChannelAccessConstants.LARGE_HEADER_SIZE);
            }
        } else {
            largeHeader = false;
            largePayloadSize = payloadSize & 0xffff;
            largeCount = count & 0xffff;
        }
        return new ChannelAccessMessageHeader(command, largePayloadSize,
                dataType, largeCount, cid, contextSpecific, largeHeader);
    }

    /**
     * Discards the next message in the byte source. If the message is larger
     * than the remaining bytes in the byte source, all bytes remaining in the
     * byte source are discarded and the number of extra bytes is returned. The
     * calling code should discard these extra bytes as soon as they are
     * received.
     * 
     * @param byteSource
     *            byte source from which the message headers are read and whose
     *            data is discarded.
     * @return the number of bytes that could not be discarded because the byte
     *         source was empty. Those bytes should be discarded when they are
     *         received, so that the next bytes that are actually processed mark
     *         the start of the next message.
     * @throws BufferUnderflowException
     *             if the byte source does not contain enough data to read the
     *             message headers. In this case no bytes are discarded at all
     *             and the byte source is left in the state it was in before
     *             calling this method.
     */
    public long discardNextMessage(final ByteSource byteSource) {
        return byteSource.atomicGet(new AtomicGetOperation<Long>() {
            @Override
            public Long get() {
                ChannelAccessMessageHeader messageHeader = decodeMessageHeader(byteSource);
                // We have to treat the payload as an unsigned 32-bit integer.
                long payloadSize = messageHeader.getPayloadSize() & 0xffffffffL;
                while (byteSource.remaining() != 0) {
                    int remaining = byteSource.remaining();
                    if (remaining < payloadSize) {
                        byteSource.skip(remaining);
                        payloadSize -= remaining;
                    } else {
                        byteSource.skip((int) payloadSize);
                        payloadSize = 0;
                        break;
                    }
                }
                return payloadSize;
            }
        });

    }

    /**
     * Completely discards the next message in the byte source. Unlike the
     * {@link #discardNextMessage(ByteSource)} method, this method will not
     * discard any bytes if the number of bytes remaining in the byte source are
     * less than the number of bytes that is expected according to the message
     * headers.
     * 
     * @param byteSource
     *            byte source from which the message headers are read and whose
     *            data is discarded.
     * @throws BufferUnderflowException
     *             if the message headers cannot be read completely or there are
     *             not enough bytes remaining in the byte source for discarding
     *             the whole message.
     */
    public void discardNextMessageCompletely(final ByteSource byteSource) {
        byteSource.atomicGet(new AtomicGetOperation<Void>() {
            @Override
            public Void get() {
                ChannelAccessMessageHeader messageHeader = decodeMessageHeader(byteSource);
                long payloadSize = messageHeader.getPayloadSize() & 0xffffffffL;
                while (payloadSize > Integer.MAX_VALUE) {
                    // We can only skip Integer.MAX_VALUE bytes at once, but the
                    // byte source might still have more data.
                    byteSource.skip(Integer.MAX_VALUE);
                    payloadSize -= Integer.MAX_VALUE;
                }
                byteSource.skip((int) payloadSize);
                return null;
            }
        });
    }

    private static ChannelAccessCommand checkMessageHeaderAndGetCommand(
            ChannelAccessMessageHeader messageHeader, int maxPayloadSize) {
        int payloadSize = messageHeader.getPayloadSize();
        if (payloadSize < 0 || payloadSize > maxPayloadSize) {
            throw new ReceivedMessageTooLargeException(
                    "Message with payload size " + (payloadSize & 0xffffffffL)
                            + " is larger than limit (" + maxPayloadSize + ").");
        }
        if (payloadSize % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT != 0) {
            throw new MalformedMessageException(
                    "Payload size of received message (" + payloadSize
                            + " bytes) is not aligned to 8 bytes.");
        }
        try {
            return ChannelAccessCommand.forCommandNumber(messageHeader
                    .getCommand());
        } catch (IllegalArgumentException e) {
            throw new UnsupportedMessageTypeException("Command number "
                    + messageHeader.getCommand()
                    + " does not refer to a well-known command.", e);
        }
    }

    private static void checkPayloadAvailable(
            ChannelAccessMessageHeader messageHeader,
            final ByteSource byteSource) {
        final int payloadSize = messageHeader.getPayloadSize();
        // If the byte source claims to have the required number of bytes, we do
        // not have to run a check.
        if (byteSource.remaining() >= payloadSize) {
            return;
        }
        // As a byte source may return a lower number of bytes than are actually
        // available, we have to try to read the number of bytes that we need in
        // order to be sure that the required data is really not available.
        try {
            byteSource.atomicGet(new AtomicGetOperation<Void>() {
                @Override
                public Void get() {
                    // Skipping is the cheapest way of testing how much data is
                    // available. For many implementations, this just means
                    // incrementing a counter. Typically, this is cheaper than
                    // actually trying to decode the message.
                    byteSource.skip(payloadSize);
                    // The required number of bytes is available. We have to
                    // throw an exception because otherwise the byte source
                    // would actually discard the data, but we still need it.
                    throw new PayloadAvailableException();
                }
            });
        } catch (BufferUnderflowException e) {
            // We transform this exception to one with information about the
            // amount of data needed.
            throw new BufferUnderflowWithSizeException(
                    messageHeader.getHeaderSize() + payloadSize);
        } catch (PayloadAvailableException e) {
            // This exception indicates success.
            return;
        }
    }

    /**
     * Decodes a message that has been sent to a server via UDP (most likely by
     * a client). The message is read from the specified byte source. This
     * method either succeeds, returning the decoded message, or fails
     * completely, leaving the byte source in its original state.
     * 
     * @param byteSource
     *            byte source from which the message is read (never
     *            <code>null</code>). The message is read in an atomic
     *            operation, thus it is either read completely or no data is
     *            read at all, leaving the byte source in its original state.
     * @param version
     *            protocol version that shall be used (never <code>null</code>).
     *            This implementation only supports protocol versions equal to
     *            or greater than {@link ChannelAccessVersion#V4_4}.
     * @return message read and decoded from the byte source (never
     *         <code>null</code>).
     * @throws BufferUnderflowException
     *             if the byte source does not contain enough data to read the
     *             whole message.
     * @throws MalformedMessageException
     *             if the message's size is not well aligned.
     * @throws ReceivedMessageTooLargeException
     *             if the message's payload is larger than the configured
     *             maximum payload receive size.
     * @throws UnsupportedMessageTypeException
     *             if the message type specified in the header is not supported
     *             by this implementation (for example because it is only
     *             expected to be received over a different transport).
     */
    public ChannelAccessMessage decodeMessageToUDPServer(
            final ByteSource byteSource, final ChannelAccessVersion version) {
        return byteSource
                .atomicGet(new AtomicGetOperation<ChannelAccessMessage>() {
                    @Override
                    public ChannelAccessMessage get() {
                        ChannelAccessMessageHeader messageHeader = decodeMessageHeader(byteSource);
                        return decodeMessageToUDPServer(messageHeader,
                                byteSource, false, charset,
                                maxPayloadReceiveSize, version);
                    }
                });
    }

    /**
     * Decodes a message that has been sent to a server via UDP (most likely by
     * a client). This method might read partial data and thus leave the byte
     * source in an undefined state. Therefore, it is only intended for internal
     * use and should only be called within an atomic get transaction. This
     * method will always read the payload specified in the message header, even
     * if the message type expects a smaller payload or no payload at all.
     * 
     * @param messageHeader
     *            headers of the message that is supposed to be decoded.
     * @param byteSource
     *            byte source from which the message's payload is read (never
     *            <code>null</code>).
     * @param headerOnly
     *            if <code>true</code>, only the message header is read. This
     *            option is useful when the message that was attached to an
     *            error message is reconstructed, because in this case only the
     *            header is available. If this option is set, no data is read
     *            from the byte source.
     * @param charset
     *            character set that is used to decode strings in the message
     *            (never <code>null</code>).
     * @param maxPayloadSize
     *            maximum size the message's payload may have. If the message's
     *            header specifies a payload size that exceeds this limit, a
     *            {@link ReceivedMessageTooLargeException} exception is thrown.
     * @param version
     *            protocol version that shall be used (never <code>null</code>).
     *            This implementation only supports protocol versions equal to
     *            or greater than {@link ChannelAccessVersion#V4_4}.
     * @return message decoded from the message header and the payload read from
     *         the byte source (never <code>null</code>).
     * @throws BufferUnderflowException
     *             if the byte source does not contain enough data to read the
     *             whole payload.
     * @throws MalformedMessageException
     *             if the message's size is not well aligned.
     * @throws ReceivedMessageTooLargeException
     *             if the message's payload is larger than
     *             <code>maxPayloadSize</code>.
     * @throws UnsupportedMessageTypeException
     *             if the message type specified in the header is not supported
     *             by this implementation (for example because it is only
     *             expected to be received over a different transport).
     */
    protected static ChannelAccessMessage decodeMessageToUDPServer(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly, Charset charset, int maxPayloadSize,
            ChannelAccessVersion version) {
        ChannelAccessCommand command;
        if (headerOnly) {
            try {
                command = ChannelAccessCommand.forCommandNumber(messageHeader
                        .getCommand());
            } catch (IllegalArgumentException e) {
                throw new UnsupportedMessageTypeException("Command number "
                        + messageHeader.getCommand()
                        + " does not refer to a well-known command.", e);
            }
        } else {
            command = checkMessageHeaderAndGetCommand(messageHeader,
                    maxPayloadSize);
        }
        switch (command) {
        case CA_PROTO_VERSION:
            return ChannelAccessVersionUDPMessage.deserialize(messageHeader,
                    byteSource, headerOnly);
        case CA_PROTO_SEARCH:
            return ChannelAccessSearchClientMessage.deserialize(messageHeader,
                    byteSource, headerOnly, charset);
        case CA_PROTO_ECHO:
            return ChannelAccessEchoMessage.deserialize(messageHeader,
                    byteSource, headerOnly);
        default:
            throw new UnsupportedMessageTypeException("A message of type "
                    + command.name() + " is not expected from a UDP client.");
        }
    }

    /**
     * Decodes a message that has been sent to a client via UDP (most likely by
     * a server). The message is read from the specified byte source. This
     * method either succeeds, returning the decoded message, or fails
     * completely, leaving the byte source in its original state.
     * 
     * @param byteSource
     *            byte source from which the message is read (never
     *            <code>null</code>). The message is read in an atomic
     *            operation, thus it is either read completely or no data is
     *            read at all, leaving the byte source in its original state.
     * @param version
     *            protocol version that shall be used (never <code>null</code>).
     *            This implementation only supports protocol versions equal to
     *            or greater than {@link ChannelAccessVersion#V4_4}.
     * @param serverAddress
     *            IP address of the server that sent the message (never
     *            <code>null</code>).
     * @param serverPort
     *            UDP port of the server that sent the message.
     * @return message read and decoded from the byte source (never
     *         <code>null</code>).
     * @throws BufferUnderflowException
     *             if the byte source does not contain enough data to read the
     *             whole message.
     * @throws MalformedMessageException
     *             if the message's size is not well aligned.
     * @throws ReceivedMessageTooLargeException
     *             if the message's payload is larger than the configured
     *             maximum payload receive size.
     * @throws UnsupportedMessageTypeException
     *             if the message type specified in the header is not supported
     *             by this implementation (for example because it is only
     *             expected to be received over a different transport).
     */
    public ChannelAccessMessage decodeMessageToUDPClient(
            final ByteSource byteSource, final ChannelAccessVersion version,
            final Inet4Address serverAddress, final int serverPort) {
        return byteSource
                .atomicGet(new AtomicGetOperation<ChannelAccessMessage>() {
                    @Override
                    public ChannelAccessMessage get() {
                        ChannelAccessMessageHeader messageHeader = decodeMessageHeader(byteSource);
                        ChannelAccessCommand command = checkMessageHeaderAndGetCommand(
                                messageHeader, maxPayloadReceiveSize);
                        switch (command) {
                        case CA_PROTO_VERSION:
                            return ChannelAccessVersionUDPMessage.deserialize(
                                    messageHeader, byteSource, false);
                        case CA_PROTO_SEARCH:
                            return ChannelAccessSearchUDPServerMessage
                                    .deserialize(messageHeader, byteSource,
                                            serverAddress, serverPort);
                        case CA_PROTO_ERROR:
                            return ChannelAccessErrorMessage.deserialize(
                                    messageHeader, byteSource, charset,
                                    version, false);
                        case CA_PROTO_NOT_FOUND:
                            return ChannelAccessNotFoundMessage.deserialize(
                                    messageHeader, byteSource);
                        case CA_PROTO_ECHO:
                            return ChannelAccessEchoMessage.deserialize(
                                    messageHeader, byteSource, false);
                        default:
                            throw new UnsupportedMessageTypeException(
                                    "A message of type "
                                            + command.name()
                                            + " is not expected from a UDP client.");
                        }
                    }
                });
    }

    /**
     * Decodes a message that has been sent from a client to a server via TCP.
     * The message is read from the specified byte source. This method either
     * succeeds, returning the decoded message, or fails completely, leaving the
     * byte source in its original state.
     * 
     * @param byteSource
     *            byte source from which the message is read (never
     *            <code>null</code>). The message is read in an atomic
     *            operation, thus it is either read completely or no data is
     *            read at all, leaving the byte source in its original state.
     * @param version
     *            protocol version that shall be used (never <code>null</code>).
     *            This implementation only supports protocol versions equal to
     *            or greater than {@link ChannelAccessVersion#V4_4}.
     * @return message read and decoded from the byte source (never
     *         <code>null</code>).
     * @throws BufferUnderflowException
     *             if the byte source does not contain enough data to read the
     *             whole message.
     * @throws MalformedMessageException
     *             if the message's size is not well aligned.
     * @throws ReceivedMessageTooLargeException
     *             if the message's payload is larger than the configured
     *             maximum payload receive size.
     * @throws ShortPayloadException
     *             if the payload size specified in the message's header is too
     *             small considering the message type and the other parameters
     *             in the message's header.
     * @throws UnsupportedMessageTypeException
     *             if the message type specified in the header is not supported
     *             by this implementation (for example because it is only
     *             expected to be received over a different transport).
     */
    public ChannelAccessMessage decodeMessageToTCPServer(
            final ByteSource byteSource, final ChannelAccessVersion version) {
        return byteSource
                .atomicGet(new AtomicGetOperation<ChannelAccessMessage>() {
                    @Override
                    public ChannelAccessMessage get() {
                        ChannelAccessMessageHeader messageHeader = decodeMessageHeader(byteSource);
                        return decodeMessageToTCPServer(messageHeader,
                                byteSource, false, charset,
                                maxPayloadReceiveSize, version);
                    }
                });
    }

    /**
     * Decodes a message that has been sent from a client to a server via TCP.
     * This method might read partial data and thus leave the byte source in an
     * undefined state. Therefore, it is only intended for internal use and
     * should only be called within an atomic get transaction. This method will
     * always read the payload specified in the message header, even if the
     * message type expects a smaller payload or no payload at all.
     * 
     * @param messageHeader
     *            headers of the message that is supposed to be decoded.
     * @param byteSource
     *            byte source from which the message's payload is read (never
     *            <code>null</code>).
     * @param headerOnly
     *            if <code>true</code>, only the message header is read. This
     *            option is useful when the message that was attached to an
     *            error message is reconstructed, because in this case only the
     *            header is available. If this option is set, no data is read
     *            from the byte source.
     * @param charset
     *            character set that is used to decode strings in the message
     *            (never <code>null</code>).
     * @param maxPayloadSize
     *            maximum size the message's payload may have. If the message's
     *            header specifies a payload size that exceeds this limit, a
     *            {@link ReceivedMessageTooLargeException} exception is thrown.
     * @param version
     *            protocol version that shall be used (never <code>null</code>).
     *            This implementation only supports protocol versions equal to
     *            or greater than {@link ChannelAccessVersion#V4_4}.
     * @return message decoded from the message header and the payload read from
     *         the byte source (never <code>null</code>).
     * @throws BufferUnderflowException
     *             if the byte source does not contain enough data to read the
     *             whole payload.
     * @throws MalformedMessageException
     *             if the message's size is not well aligned.
     * @throws ReceivedMessageTooLargeException
     *             if the message's payload is larger than
     *             <code>maxPayloadSize</code>.
     * @throws ShortPayloadException
     *             if the payload size specified in the message's header is too
     *             small considering the message type and the other parameters
     *             in the message's header.
     * @throws UnsupportedMessageTypeException
     *             if the message type specified in the header is not supported
     *             by this implementation (for example because it is only
     *             expected to be received over a different transport).
     */
    protected static ChannelAccessMessage decodeMessageToTCPServer(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly, Charset charset, int maxPayloadSize,
            ChannelAccessVersion version) {
        ChannelAccessCommand command;
        if (headerOnly) {
            try {
                command = ChannelAccessCommand.forCommandNumber(messageHeader
                        .getCommand());
            } catch (IllegalArgumentException e) {
                throw new UnsupportedMessageTypeException("Command number "
                        + messageHeader.getCommand()
                        + " does not refer to a well-known command.", e);
            }
        } else {
            command = checkMessageHeaderAndGetCommand(messageHeader,
                    maxPayloadSize);
            // A TCP message can be rather large and therefore we want to test
            // whether there is enough data available for the whole payload
            // before actually trying to decode it. This also has the advantage
            // that we will throw a BufferUnderflowWithSizeException, so that
            // the calling code knows how much data is missing. We already
            // checked the message header, so we know that the payload size is
            // within the configured limits.
            checkPayloadAvailable(messageHeader, byteSource);
        }
        switch (command) {
        case CA_PROTO_VERSION:
            return ChannelAccessVersionTCPClientMessage.deserialize(
                    messageHeader, byteSource, headerOnly);
        case CA_PROTO_CLIENT_NAME:
            return ChannelAccessUserNameMessage.deserialize(messageHeader,
                    byteSource, headerOnly, charset);
        case CA_PROTO_HOST_NAME:
            return ChannelAccessHostNameMessage.deserialize(messageHeader,
                    byteSource, headerOnly, charset);
        case CA_PROTO_SEARCH:
            return ChannelAccessSearchClientMessage.deserialize(messageHeader,
                    byteSource, headerOnly, charset);
        case CA_PROTO_CREATE_CHAN:
            return ChannelAccessConnectChannelClientMessage.deserialize(
                    messageHeader, byteSource, headerOnly, charset);
        case CA_PROTO_CLEAR_CHANNEL:
            return ChannelAccessDisconnectChannelMessage.deserialize(
                    messageHeader, byteSource, headerOnly);
        case CA_PROTO_EVENT_ADD:
            return ChannelAccessSubscriptionClientMessage.deserialize(
                    messageHeader, byteSource, headerOnly);
        case CA_PROTO_EVENT_CANCEL:
            return ChannelAccessCancelSubscriptionClientMessage.deserialize(
                    messageHeader, byteSource, headerOnly);
        case CA_PROTO_READ:
            return ChannelAccessReadClientMessage.deserialize(messageHeader,
                    byteSource, headerOnly);
        case CA_PROTO_WRITE:
            return ChannelAccessWriteMessage.deserialize(messageHeader,
                    byteSource, headerOnly, charset);
        case CA_PROTO_READ_NOTIFY:
            return ChannelAccessReadNotifyClientMessage.deserialize(
                    messageHeader, byteSource, headerOnly);
        case CA_PROTO_WRITE_NOTIFY:
            return ChannelAccessWriteNotifyClientMessage.deserialize(
                    messageHeader, byteSource, headerOnly, charset);
        case CA_PROTO_ECHO:
            return ChannelAccessEchoMessage.deserialize(messageHeader,
                    byteSource, headerOnly);
        case CA_PROTO_EVENTS_OFF:
            return ChannelAccessEventsOffMessage.deserialize(messageHeader,
                    byteSource, headerOnly);
        case CA_PROTO_EVENTS_ON:
            return ChannelAccessEventsOnMessage.deserialize(messageHeader,
                    byteSource, headerOnly);
        case CA_PROTO_READ_SYNC:
            return ChannelAccessReadSyncMessage.deserialize(messageHeader,
                    byteSource, headerOnly);
        case CA_PROTO_READ_BUILD:
            // Some old clients might send this message. The message object that
            // we construct here is just a dummy, so that the calling code can
            // identify and ignore this message.
            return ChannelAccessReadBuildMessage.deserialize(messageHeader,
                    byteSource, headerOnly);
        default:
            throw new UnsupportedMessageTypeException("A message of type "
                    + command.name() + " is not expected from a TCP client.");
        }
    }

    /**
     * Decodes a message that has been sent from a server to a client via TCP.
     * The message is read from the specified byte source. This method either
     * succeeds, returning the decoded message, or fails completely, leaving the
     * byte source in its original state.
     * 
     * @param byteSource
     *            byte source from which the message is read (never
     *            <code>null</code>). The message is read in an atomic
     *            operation, thus it is either read completely or no data is
     *            read at all, leaving the byte source in its original state.
     * @param version
     *            protocol version that shall be used (never <code>null</code>).
     *            This implementation only supports protocol versions equal to
     *            or greater than {@link ChannelAccessVersion#V4_4}.
     * @param serverAddress
     *            IP address of the server that sent the message (never
     *            <code>null</code>).
     * @param serverPort
     *            TCP port of the server that sent the message.
     * @return message read and decoded from the byte source (never
     *         <code>null</code>).
     * @throws BufferUnderflowException
     *             if the byte source does not contain enough data to read the
     *             whole message.
     * @throws MalformedMessageException
     *             if the message's size is not well aligned.
     * @throws ReceivedMessageTooLargeException
     *             if the message's payload is larger than the configured
     *             maximum payload receive size.
     * @throws ShortPayloadException
     *             if the payload size specified in the message's header is too
     *             small considering the message type and the other parameters
     *             in the message's header.
     * @throws UnsupportedMessageTypeException
     *             if the message type specified in the header is not supported
     *             by this implementation (for example because it is only
     *             expected to be received over a different transport).
     */
    public ChannelAccessMessage decodeMessageToTCPClient(
            final ByteSource byteSource, final ChannelAccessVersion version,
            final Inet4Address serverAddress, final int serverPort) {
        return byteSource
                .atomicGet(new AtomicGetOperation<ChannelAccessMessage>() {
                    @Override
                    public ChannelAccessMessage get() {
                        ChannelAccessMessageHeader messageHeader = decodeMessageHeader(byteSource);
                        // A TCP message can be rather large and therefore we
                        // want to test whether there is enough data available
                        // for the whole payload before actually trying to
                        // decode it. This also has the advantage that we will
                        // throw a BufferUnderflowWithSizeException, so that the
                        // calling code knows how much data is missing. We
                        // already checked the message header, so we know that
                        // the payload size is within the configured limits.
                        checkPayloadAvailable(messageHeader, byteSource);
                        ChannelAccessCommand command = checkMessageHeaderAndGetCommand(
                                messageHeader, maxPayloadReceiveSize);
                        switch (command) {
                        case CA_PROTO_VERSION:
                            return ChannelAccessVersionTCPServerMessage
                                    .deserialize(messageHeader, byteSource);
                        case CA_PROTO_SEARCH:
                            return ChannelAccessSearchTCPServerMessage
                                    .deserialize(messageHeader, byteSource,
                                            serverAddress, serverPort);
                        case CA_PROTO_NOT_FOUND:
                            return ChannelAccessNotFoundMessage.deserialize(
                                    messageHeader, byteSource);
                        case CA_PROTO_ACCESS_RIGHTS:
                            return ChannelAccessAccessRightsMessage
                                    .deserialize(messageHeader, byteSource);
                        case CA_PROTO_CREATE_CHAN:
                            return ChannelAccessConnectChannelServerMessage
                                    .deserialize(messageHeader, byteSource);
                        case CA_PROTO_CLEAR_CHANNEL:
                            return ChannelAccessDisconnectChannelMessage
                                    .deserialize(messageHeader, byteSource,
                                            false);
                        case CA_PROTO_CREATE_CH_FAIL:
                            return ChannelAccessConnectChannelFailedMessage
                                    .deserialize(messageHeader, byteSource);
                        case CA_PROTO_SERVER_DISCONN:
                            return ChannelAccessChannelDisconnectedByServerMessage
                                    .deserialize(messageHeader, byteSource);
                        case CA_PROTO_EVENT_ADD:
                            if (messageHeader.getPayloadSize() == 0) {
                                return ChannelAccessCancelSubscriptionServerMessage
                                        .deserialize(messageHeader, byteSource);
                            } else {
                                return ChannelAccessSubscriptionServerMessage
                                        .deserialize(messageHeader, byteSource,
                                                charset);
                            }
                        case CA_PROTO_READ:
                            return ChannelAccessReadServerMessage.deserialize(
                                    messageHeader, byteSource, charset);
                        case CA_PROTO_READ_NOTIFY:
                            return ChannelAccessReadNotifyServerMessage
                                    .deserialize(messageHeader, byteSource,
                                            charset);
                        case CA_PROTO_WRITE_NOTIFY:
                            return ChannelAccessWriteNotifyServerMessage
                                    .deserialize(messageHeader, byteSource);
                        case CA_PROTO_ERROR:
                            return ChannelAccessErrorMessage.deserialize(
                                    messageHeader, byteSource, charset,
                                    version, true);
                        case CA_PROTO_ECHO:
                            return ChannelAccessEchoMessage.deserialize(
                                    messageHeader, byteSource, false);
                        case CA_PROTO_READ_SYNC:
                            return ChannelAccessReadSyncMessage.deserialize(
                                    messageHeader, byteSource, false);
                        default:
                            throw new UnsupportedMessageTypeException(
                                    "A message of type "
                                            + command.name()
                                            + " is not expected from a UDP client.");
                        }
                    }
                });
    }

    /**
     * Decodes a message that has been sent to a repeater via UDP (from a client
     * or a server). The message is read from the specified byte source. This
     * method either succeeds, returning the decoded message, or fails
     * completely, leaving the byte source in its original state.
     * 
     * @param byteSource
     *            byte source from which the message is read (never
     *            <code>null</code>). The message is read in an atomic
     *            operation, thus it is either read completely or no data is
     *            read at all, leaving the byte source in its original state.
     * @param version
     *            protocol version that shall be used (never <code>null</code>).
     *            This implementation only supports protocol versions equal to
     *            or greater than {@link ChannelAccessVersion#V4_4}.
     * @param originAddress
     *            IP address of the host that the message originated from.
     * @return message read and decoded from the byte source (never
     *         <code>null</code>).
     * @throws BufferUnderflowException
     *             if the byte source does not contain enough data to read the
     *             message header.
     * @throws MalformedMessageException
     *             if the message's size is not well aligned.
     * @throws ReceivedMessageTooLargeException
     *             if the message's payload is larger than the configured
     *             maximum payload receive size.
     * @throws UnsupportedMessageTypeException
     *             if the message type specified in the header is not supported
     *             by this implementation (for example because it is only
     *             expected to be received over a different transport).
     */
    public ChannelAccessMessage decodeMessageToRepeater(
            final ByteSource byteSource, final ChannelAccessVersion version,
            final Inet4Address originAddress) {
        return byteSource
                .atomicGet(new AtomicGetOperation<ChannelAccessMessage>() {
                    @Override
                    public ChannelAccessMessage get() {
                        ChannelAccessMessageHeader messageHeader = decodeMessageHeader(byteSource);
                        ChannelAccessCommand command = checkMessageHeaderAndGetCommand(
                                messageHeader, maxPayloadReceiveSize);
                        switch (command) {
                        case CA_PROTO_RSRV_IS_UP:
                            return ChannelAccessBeaconMessage.deserialize(
                                    messageHeader, byteSource, originAddress);
                        case REPEATER_REGISTER:
                            return ChannelAccessRepeaterRegisterMessage
                                    .deserialize(messageHeader, byteSource);
                        default:
                            throw new UnsupportedMessageTypeException(
                                    "A message of type " + command.name()
                                            + " is not expected by a repeater.");
                        }
                    }
                });
    }

    /**
     * Decodes a message that has been sent from a repeater to a client via UDP.
     * The message is read from the specified byte source. This method either
     * succeeds, returning the decoded message, or fails completely, leaving the
     * byte source in its original state.
     * 
     * @param byteSource
     *            byte source from which the message is read (never
     *            <code>null</code>). The message is read in an atomic
     *            operation, thus it is either read completely or no data is
     *            read at all, leaving the byte source in its original state.
     * @param version
     *            protocol version that shall be used (never <code>null</code>).
     *            This implementation only supports protocol versions equal to
     *            or greater than {@link ChannelAccessVersion#V4_4}.
     * @return message read and decoded from the byte source (never
     *         <code>null</code>).
     * @throws BufferUnderflowException
     *             if the byte source does not contain enough data to read the
     *             message header.
     * @throws MalformedMessageException
     *             if the message's size is not well aligned.
     * @throws ReceivedMessageTooLargeException
     *             if the message's payload is larger than the configured
     *             maximum payload receive size.
     * @throws UnsupportedMessageTypeException
     *             if the message type specified in the header is not supported
     *             by this implementation (for example because it is only
     *             expected to be received over a different transport).
     */
    public ChannelAccessMessage decodeMessageFromRepeater(
            final ByteSource byteSource, final ChannelAccessVersion version) {
        return byteSource
                .atomicGet(new AtomicGetOperation<ChannelAccessMessage>() {
                    @Override
                    public ChannelAccessMessage get() {
                        ChannelAccessMessageHeader messageHeader = decodeMessageHeader(byteSource);
                        ChannelAccessCommand command = checkMessageHeaderAndGetCommand(
                                messageHeader, maxPayloadReceiveSize);
                        switch (command) {
                        case CA_PROTO_VERSION:
                            // The repeater sends out a dummy version message
                            // when testing whether clients are still alive.
                            return ChannelAccessVersionUDPMessage.deserialize(
                                    messageHeader, byteSource, false);
                        case CA_PROTO_RSRV_IS_UP:
                            return ChannelAccessBeaconMessage.deserialize(
                                    messageHeader, byteSource, null);
                        case REPEATER_CONFIRM:
                            return ChannelAccessRepeaterConfirmMessage
                                    .deserialize(messageHeader, byteSource);
                        default:
                            throw new UnsupportedMessageTypeException(
                                    "A message of type "
                                            + command.name()
                                            + " is not expected from a repeater.");
                        }
                    }
                });
    }

    private void verifyMessageInternal(ChannelAccessMessage message,
            ChannelAccessVersion version) {
        message.verify(version, maxPayloadSendSize, charset);
    }

    /**
     * Verifies a message that shall be sent from a client to a server via UDP.
     * Only messages of type {@link ChannelAccessVersionUDPMessage} and
     * {@link ChannelAccessSearchClientMessage} can be sent via this protocol.
     * This verification ensures that
     * {@link #encodeMessageToUDPServer(ByteSink, ChannelAccessMessage, ChannelAccessVersion)}
     * does not throw an exception (except for a {@link BufferOverflowException}
     * ).
     * 
     * @param message
     *            message to be verified.
     * @param version
     *            protocol version that shall be used. This implementation only
     *            supports protocol versions equal to or greater than
     *            {@link ChannelAccessVersion#V4_4}.
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent from a
     *             client to a server via UDP.
     * @throws SerializedChannelNameTooLargeException
     *             if a channel name has to be serialized and the serialized
     *             form of the channel name is too long.
     * @throws SerializedPayloadTooLargeException
     *             if the message cannot be sent because the payload would be
     *             larger than the configured maximum payload send size or the
     *             specified protocol version is older than Channel Access
     *             version 4.9 and the payload would be larger than 65528 bytes
     *             or the count exceeds 65535 elements.
     * @throws UnsupportedInThisVersionException
     *             if this message is not support in the protocol version
     *             specified by <code>version</code>.
     * @throws UnsupportedOperationException
     *             if this message has been constructed by deserializing it from
     *             headers only and thus it is incomplete.
     * @throws UnsupportedProtocolVersionException
     *             if the specified <code>version</code> is older than the
     *             oldest supported protocol version for this message. In
     *             general, all Channel Access versions since version 4.4 should
     *             be supported.
     */
    public void verifyMessageToUDPServer(ChannelAccessMessage message,
            ChannelAccessVersion version) {
        if (message instanceof ChannelAccessVersionUDPMessage
                || message instanceof ChannelAccessSearchClientMessage) {
            verifyMessageInternal(message, version);
        } else {
            // If we got an unsupported message type for this type of transport,
            // this must be a bug in the implementation, therefore throwing an
            // IllegalArgumentException seems appropriate.
            throw new IllegalArgumentException("Cannot send message of type "
                    + message.getClass().getName() + " to UDP server.");
        }
    }

    /**
     * Verifies a message that shall be sent from a server to a client via UDP.
     * Only messages of type {@link ChannelAccessVersionUDPMessage} and
     * {@link ChannelAccessSearchUDPServerMessage},
     * {@link ChannelAccessEchoMessage}, {@link ChannelAccessErrorMessage}, and
     * {@link ChannelAccessNotFoundMessage} can be sent via this protocol. This
     * verification ensures that
     * {@link #encodeMessageToUDPClient(ByteSink, ChannelAccessMessage, ChannelAccessVersion)}
     * does not throw an exception (except for a {@link BufferOverflowException}
     * ).
     * 
     * @param message
     *            message to be encoded.
     * @param version
     *            protocol version that shall be used. This implementation only
     *            supports protocol versions equal to or greater than
     *            {@link ChannelAccessVersion#V4_4}.
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent from a
     *             server to a client via UDP.
     * @throws SerializedPayloadTooLargeException
     *             if the message cannot be sent because the payload would be
     *             larger than the configured maximum payload send size or the
     *             specified protocol version is older than Channel Access
     *             version 4.9 and the payload would be larger than 65528 bytes
     *             or the count exceeds 65535 elements.
     * @throws UnsupportedInThisVersionException
     *             if this message is not support in the protocol version
     *             specified by <code>version</code>.
     * @throws UnsupportedOperationException
     *             if this message has been constructed by deserializing it from
     *             headers only and thus it is incomplete.
     * @throws UnsupportedProtocolVersionException
     *             if the specified <code>version</code> is older than the
     *             oldest supported protocol version for this message. In
     *             general, all Channel Access versions since version 4.4 should
     *             be supported.
     */
    public void verifyMessageToUDPClient(ChannelAccessMessage message,
            ChannelAccessVersion version) {
        if (message instanceof ChannelAccessVersionUDPMessage
                || message instanceof ChannelAccessSearchUDPServerMessage
                || message instanceof ChannelAccessEchoMessage
                || message instanceof ChannelAccessErrorMessage
                || message instanceof ChannelAccessNotFoundMessage) {
            verifyMessageInternal(message, version);
        } else {
            // If we got an unsupported message type for this type of transport,
            // this must be a bug in the implementation, therefore throwing an
            // IllegalArgumentException seems appropriate.
            throw new IllegalArgumentException("Cannot send message of type "
                    + message.getClass().getName() + " to UDP client.");
        }
    }

    /**
     * Verifies a message that shall be sent from a server or a client to a
     * repeater via UDP. Only messages of type
     * {@link ChannelAccessBeaconMessage} can be sent via this protocol. This
     * verification ensures that
     * {@link #encodeMessageToRepeater(ByteSink, ChannelAccessMessage, ChannelAccessVersion)}
     * does not throw an exception (except for a {@link BufferOverflowException}
     * ).
     * 
     * @param message
     *            message to be encoded.
     * @param version
     *            protocol version that shall be used. This implementation only
     *            supports protocol versions equal to or greater than
     *            {@link ChannelAccessVersion#V4_4}.
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent to a
     *             repeater.
     */
    public void verifyMessageToRepeater(ChannelAccessMessage message,
            ChannelAccessVersion version) {
        // We could also support ChannelAccessRepeaterRegisterMessage here, but
        // this would just result in an UnsupportedOperationException when we
        // call the serialize method.
        if (message instanceof ChannelAccessBeaconMessage) {
            verifyMessageInternal(message, version);
        } else {
            // If we got an unsupported message type for this type of transport,
            // this must be a bug in the implementation, therefore throwing an
            // IllegalArgumentException seems appropriate.
            throw new IllegalArgumentException("Cannot send message of type "
                    + message.getClass().getName() + " to repeater.");
        }
    }

    /**
     * Verifies a message that shall be sent from a repeater to a client via
     * UDP. Only messages of type {@link ChannelAccessVersionUDPMessage},
     * {@link ChannelAccessRepeaterConfirmMessage}, and
     * {@link ChannelAccessBeaconMessage} can be sent via this protocol. This
     * verification ensures that
     * {@link #encodeMessageFromRepeater(ByteSink, ChannelAccessMessage, ChannelAccessVersion)}
     * does not throw an exception (except for a {@link BufferOverflowException}
     * ).
     * 
     * @param message
     *            message to be encoded.
     * @param version
     *            protocol version that shall be used. This implementation only
     *            supports protocol versions equal to or greater than
     *            {@link ChannelAccessVersion#V4_4}.
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent from a
     *             repeater to a client via UDP.
     */
    public void verifyMessageFromRepeater(ChannelAccessMessage message,
            ChannelAccessVersion version) {
        if (message instanceof ChannelAccessVersionUDPMessage
                || message instanceof ChannelAccessRepeaterConfirmMessage
                || message instanceof ChannelAccessBeaconMessage) {
            verifyMessageInternal(message, version);
        } else {
            // If we got an unsupported message type for this type of transport,
            // this must be a bug in the implementation, therefore throwing an
            // IllegalArgumentException seems appropriate.
            throw new IllegalArgumentException("Cannot send message of type "
                    + message.getClass().getName() + " from repeater.");
        }
    }

    /**
     * Verifies a message that shall be sent from a client to a server via TCP
     * Only messages of type {@link ChannelAccessVersionTCPClientMessage},
     * {@link ChannelAccessSubscriptionClientMessage},
     * {@link ChannelAccessCancelSubscriptionClientMessage},
     * {@link ChannelAccessReadClientMessage}, {@link ChannelAccessWriteMessage}
     * , {@link ChannelAccessEventsOffMessage},
     * {@link ChannelAccessEventsOnMessage},
     * {@link ChannelAccessReadSyncMessage},
     * {@link ChannelAccessDisconnectChannelMessage},
     * {@link ChannelAccessReadNotifyClientMessage},
     * {@link ChannelAccessConnectChannelClientMessage},
     * {@link ChannelAccessWriteNotifyClientMessage},
     * {@link ChannelAccessUserNameMessage},
     * {@link ChannelAccessHostNameMessage}, {@link ChannelAccessEchoMessage},
     * and {@link ChannelAccessSearchClientMessage} can be sent via this
     * protocol. This verification ensures that
     * {@link #encodeMessageToTCPServer(ByteSink, ChannelAccessMessage, ChannelAccessVersion)}
     * does not throw an exception (except for a {@link BufferOverflowException}
     * ).
     * 
     * @param message
     *            message to be encoded.
     * @param version
     *            protocol version that shall be used. This implementation only
     *            supports protocol versions equal to or greater than
     *            {@link ChannelAccessVersion#V4_4}.
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent from a
     *             client to a server via TCP.
     * @throws SerializedPayloadTooLargeException
     *             if the message cannot be sent because the payload would be
     *             larger than the configured maximum payload send size or the
     *             specified protocol version is older than Channel Access
     *             version 4.9 and the payload would be larger than 65528 bytes
     *             or the count exceeds 65535 elements.
     * @throws UnsupportedInThisVersionException
     *             if this message is not support in the protocol version
     *             specified by <code>version</code>.
     * @throws UnsupportedOperationException
     *             if this message has been constructed by deserializing it from
     *             headers only and thus it is incomplete.
     * @throws UnsupportedProtocolVersionException
     *             if the specified <code>version</code> is older than the
     *             oldest supported protocol version for this message. In
     *             general, all Channel Access versions since version 4.4 should
     *             be supported.
     */
    public void verifyMessageToTCPServer(ChannelAccessMessage message,
            ChannelAccessVersion version) {
        // We could also support ChannelAccessReadBuildMessage here, but this
        // would just result in an UnsupportedOperationException when we call
        // the serialize method.
        if (message instanceof ChannelAccessVersionTCPClientMessage
                || message instanceof ChannelAccessSubscriptionClientMessage
                || message instanceof ChannelAccessCancelSubscriptionClientMessage
                || message instanceof ChannelAccessReadClientMessage
                || message instanceof ChannelAccessWriteMessage
                || message instanceof ChannelAccessEventsOffMessage
                || message instanceof ChannelAccessEventsOnMessage
                || message instanceof ChannelAccessReadSyncMessage
                || message instanceof ChannelAccessDisconnectChannelMessage
                || message instanceof ChannelAccessReadNotifyClientMessage
                || message instanceof ChannelAccessConnectChannelClientMessage
                || message instanceof ChannelAccessWriteNotifyClientMessage
                || message instanceof ChannelAccessUserNameMessage
                || message instanceof ChannelAccessHostNameMessage
                || message instanceof ChannelAccessEchoMessage
                || message instanceof ChannelAccessSearchClientMessage) {
            verifyMessageInternal(message, version);
        } else {
            // If we got an unsupported message type for this type of transport,
            // this must be a bug in the implementation, therefore throwing an
            // IllegalArgumentException seems appropriate.
            throw new IllegalArgumentException("Cannot send message of type "
                    + message.getClass().getName() + " to TCP server.");
        }
    }

    /**
     * Verifies a message that shall be sent from a server to a client via TCP
     * Only messages of type {@link ChannelAccessVersionTCPServerMessage},
     * {@link ChannelAccessSubscriptionServerMessage},
     * {@link ChannelAccessCancelSubscriptionServerMessage},
     * {@link ChannelAccessReadServerMessage},
     * {@link ChannelAccessReadSyncMessage} , {@link ChannelAccessErrorMessage},
     * {@link ChannelAccessDisconnectChannelMessage},
     * {@link ChannelAccessReadNotifyServerMessage},
     * {@link ChannelAccessConnectChannelServerMessage},
     * {@link ChannelAccessWriteNotifyServerMessage},
     * {@link ChannelAccessAccessRightsMessage},
     * {@link ChannelAccessEchoMessage},
     * {@link ChannelAccessConnectChannelFailedMessage},
     * {@link ChannelAccessChannelDisconnectedByServerMessage},
     * {@link ChannelAccessSearchTCPServerMessage}, and
     * {@link ChannelAccessNotFoundMessage} can be sent via this protocol. This
     * verification ensures that
     * {@link #encodeMessageToTCPClient(ByteSink, ChannelAccessMessage, ChannelAccessVersion)}
     * does not throw an exception (except for a {@link BufferOverflowException}
     * ).
     * 
     * @param message
     *            message to be encoded.
     * @param version
     *            protocol version that shall be used. This implementation only
     *            supports protocol versions equal to or greater than
     *            {@link ChannelAccessVersion#V4_4}.
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent from a
     *             server to a client via TCP.
     * @throws SerializedPayloadTooLargeException
     *             if the message cannot be sent because the payload would be
     *             larger than the configured maximum payload send size or the
     *             specified protocol version is older than Channel Access
     *             version 4.9 and the payload would be larger than 65528 bytes
     *             or the count exceeds 65535 elements.
     * @throws UnsupportedInThisVersionException
     *             if this message is not support in the protocol version
     *             specified by <code>version</code>.
     * @throws UnsupportedOperationException
     *             if this message has been constructed by deserializing it from
     *             headers only and thus it is incomplete.
     * @throws UnsupportedProtocolVersionException
     *             if the specified <code>version</code> is older than the
     *             oldest supported protocol version for this message. In
     *             general, all Channel Access versions since version 4.4 should
     *             be supported.
     */
    public void verifyMessageToTCPClient(ChannelAccessMessage message,
            ChannelAccessVersion version) {
        // We could also support ChannelAccessReadBuildMessage here, but this
        // would just result in an UnsupportedOperationException when we call
        // the serialize method.
        if (message instanceof ChannelAccessVersionTCPServerMessage
                || message instanceof ChannelAccessSubscriptionServerMessage
                || message instanceof ChannelAccessCancelSubscriptionServerMessage
                || message instanceof ChannelAccessReadServerMessage
                || message instanceof ChannelAccessReadSyncMessage
                || message instanceof ChannelAccessErrorMessage
                || message instanceof ChannelAccessDisconnectChannelMessage
                || message instanceof ChannelAccessReadNotifyServerMessage
                || message instanceof ChannelAccessConnectChannelServerMessage
                || message instanceof ChannelAccessWriteNotifyServerMessage
                || message instanceof ChannelAccessAccessRightsMessage
                || message instanceof ChannelAccessEchoMessage
                || message instanceof ChannelAccessConnectChannelFailedMessage
                || message instanceof ChannelAccessChannelDisconnectedByServerMessage
                || message instanceof ChannelAccessSearchTCPServerMessage
                || message instanceof ChannelAccessNotFoundMessage) {
            verifyMessageInternal(message, version);
        } else {
            // If we got an unsupported message type for this type of transport,
            // this must be a bug in the implementation, therefore throwing an
            // IllegalArgumentException seems appropriate.
            throw new IllegalArgumentException("Cannot send message of type "
                    + message.getClass().getName() + " to TCP client.");
        }
    }

}
