/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.io.IOException;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.SystemUtils;

/**
 * Host name utility. Provides a method for determining the host name of the
 * local machine.
 * 
 * @author Sebastian Marsching
 */
public final class HostNameUtil {

    private HostNameUtil() {
    }

    /**
     * Returns the host name of the local machine. This method first tries to
     * determine the host name by running the "hostname" command. If this is not
     * successful, the "HOSTNAME" (for UNIX-like systems) and "COMPUTERNAME"
     * (for Windows) environment variables are tried. If the respective
     * environment variable is not defined either,
     * {@link InetAddress#getLocalHost()} is tried. If that method does not
     * return an address with a valid host name or throws an exception,
     * <code>null</code> is returned.
     * 
     * @return host name of the local machine or <code>null</code> if the host
     *         name cannot be determined.
     */
    public static String getHostName() {
        String hostName;
        // First we try to run the "hostname" command. On most platforms, this
        // will work without DNS and will return the same value that the
        // gethostname system call returns.
        hostName = hostNameFromCommand();
        if (hostName != null) {
            return hostName;
        }
        // If the "hostname" command cannot be found or there is some other
        // problem with running it, we try to get the host name from environment
        // variables. They might not always be defined and may contain a
        // different value, so they are our second choice.
        if (SystemUtils.IS_OS_UNIX) {
            hostName = System.getenv("HOSTNAME");
        } else if (SystemUtils.IS_OS_WINDOWS) {
            hostName = System.getenv("COMPUTERNAME");
        }
        if (hostName != null) {
            return hostName;
        }
        // Finally, we try InetAddress.getLocalHost(). This will only work if
        // DNS is setup correctly and might return a different value than
        // expected. Therefore, we try this variant last.
        try {
            hostName = InetAddress.getLocalHost().getHostName();
            int firstDot = hostName.indexOf('.');
            if (firstDot == -1) {
                return hostName;
            } else {
                hostName = hostName.substring(0, firstDot);
                // If the host name is all numeric, we probably got an IP
                // address.
                if (!StringUtils.isNumeric(hostName)) {
                    return hostName;
                }
            }
            hostName = null;
        } catch (UnknownHostException e) {
            // Ignore the exception, we simply continue with a null value.
        }
        // If everything failed, we return null.
        return null;
    }

    private static String hostNameFromCommand() {
        // Run the "hostname" command to get the hostname. This should work on
        // Windows and on UNIX-like systems.
        final Process process;
        try {
            process = new ProcessBuilder("hostname").start();
        } catch (IOException e) {
            return null;
        }
        // We read from the process in a separate thread. This way, we have a
        // chance to kill the process if - for some reason - it does not quit as
        // expected.
        FutureTask<String> hostnameTask = new FutureTask<String>(
                new Callable<String>() {
                    @Override
                    public String call() throws Exception {
                        StringBuilder sb = new StringBuilder();
                        InputStreamReader reader = new InputStreamReader(
                                process.getInputStream());
                        int nextChar;
                        while ((nextChar = reader.read()) != -1) {
                            sb.append((char) nextChar);
                        }
                        return sb.toString();
                    }
                });
        Thread readThread = new Thread(hostnameTask);
        readThread.setDaemon(true);
        readThread.start();
        String hostName = null;
        try {
            // We wait for up to one second. If the "hostname" command does not
            // finish within this time, there clearly is a problem and we better
            // kill the process.
            hostName = hostnameTask.get(1000L, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
        } catch (ExecutionException e) {
            // Ignore
        } catch (TimeoutException e) {
            // Ignore
        }
        // If the hostname process did not quit properly, the read thread might
        // wait forever. Therefore, we kill the process if it is still running.
        try {
            // If the process quit with a non-zero return code, the output might
            // be unusable.
            if (process.exitValue() == 0) {
                return hostName.trim();
            } else {
                return null;
            }
        } catch (IllegalThreadStateException e) {
            // Ignore, this just means that the process is still running.
        }
        process.destroy();
        return null;
    }

}
