/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.util.concurrent.Executor;
import java.util.concurrent.Future;

import com.aquenos.epics.jackie.common.exception.ConcurrentNotificationException;

/**
 * <p>
 * Future that allows asynchronous notifications.
 * </p>
 * 
 * <p>
 * This kind of future can be used to be notified when the result is available,
 * thus removing the need for a blocking thread that waits on the result.
 * </p>
 * 
 * @author Sebastian Marsching
 *
 * @param <V>
 *            The result type returned by the future.
 */
public interface ListenableFuture<V> extends Future<V> {

    /**
     * <p>
     * Registers a completion listener with this future. This listener is
     * notified when the future completes ({@link #isDone()} returns
     * <code>true</code>). If the future completes before the listener is
     * registered, the listener is notified immediately while being added.
     * </p>
     * 
     * <p>
     * If the specified listener is already registered with this future, no
     * action is taken and this method returns <code>false</code>.
     * </p>
     * 
     * <p>
     * <strong>Important note</strong>: Typically, the listener is executed by
     * the thread that sets the future's value. For this reason, it is important
     * that the listener does not block or perform any long-running operations.
     * Instead, the listener should delegate any such operation to a different
     * thread (e.g. using an {@link Executor}). Any violation of this rule might
     * result in unexpected or unpredictable behavior. In particular, a blocking
     * listener might cause a dead-lock, which results in the whole application
     * being stalled indefinitely.
     * </p>
     * 
     * @param listener
     *            listener to be registered with this future.
     * @return <code>true</code> if the listener has been added,
     *         <code>false</code> if it has already been added earlier.
     */
    boolean addCompletionListener(FutureCompletionListener<? super V> listener);

    /**
     * Returns the listener lock-policy used by this future. The listener
     * lock-policy defines the behavior of the
     * {@link #removeCompletionListener(FutureCompletionListener)} method when a
     * listener is removed while a notification of the same listener is in
     * progress.
     * 
     * @return listener lock-policy used by this future.
     * @since EPICS Jackie 2.0.0
     */
    ListenerLockPolicy getListenerLockPolicy();

    /**
     * <p>
     * Removes the registration of a completion listener with this future. The
     * exact behavior of this method depends on the
     * {@linkplain #getListenerLockPolicy() listener lock-policy} used by this
     * future.
     * </p>
     * 
     * <p>
     * If the {@link ListenerLockPolicy#BLOCK BLOCK} policy is in effect, the
     * listener is not going to receive a completion notification after this
     * method returns. If the {@link ListenerLockPolicy#IGNORE IGNORE} policy is
     * in effect, the listener might still be notified after this method
     * returns. This may only happen if the future completes before this method
     * returns. If the {@link ListenerLockPolicy#REPORT REPORT} policy is in
     * effect, this method throws a {@link ConcurrentNotificationException} if
     * the listener might still be notified. The listener is still removed,
     * despite the exception being thrown.
     * </p>
     * 
     * <p>
     * If the specified listener is not currently registered with this future,
     * no action is taken and this method returns <code>false</code>.
     * </p>
     * 
     * @param listener
     *            listener to be unregistered from this future.
     * @return <code>true</code> if the listener has been removed,
     *         <code>false</code> if it has already been removed earlier of if
     *         has never been added.
     * @throws ConcurrentNotificationException
     *             if the {@link ListenerLockPolicy} is set to {@code REPORT}
     *             and the removed listener might still receive a notification
     *             after this method has returned. The listener is still
     *             removed, even if this exception is thrown.
     * @see #getListenerLockPolicy()
     */
    boolean removeCompletionListener(
            FutureCompletionListener<? super V> listener);

}
