/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.util.ListIterator;
import java.util.NoSuchElementException;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.value.ChannelAccessAcknowledgeAlarm;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmAcknowledgementStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessConfigureAcknowledgeTransientAlarms;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessFloatingPointGraphicsValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessNumericControlsValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessNumericGraphicsValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessValue;

/**
 * Base implementation for Channel Access values.
 * 
 * @param <ElementType>
 *            type of the actual value elements. For primitives, this is the
 *            boxed variant.
 * @param <ImplementationType>
 *            type of the actual implementation.
 * 
 * @author Sebastian Marsching
 */
public abstract class ChannelAccessValueBase<ElementType, ImplementationType extends ChannelAccessValueBase<ElementType, ImplementationType>>
        implements ChannelAccessValue<ElementType> {

    /**
     * Delegate to use if this value is a read-only value. If this is not
     * <code>null</code>, all read operations should actually work on this value
     * and write operations should not be allowed.
     */
    protected ImplementationType delegate;

    /**
     * Default base constructor.
     */
    ChannelAccessValueBase() {
    }

    /**
     * Creates a read-only value.
     * 
     * @param delegate
     *            delegate value all read operations are delegated too.
     */
    ChannelAccessValueBase(ImplementationType delegate) {
        this.delegate = delegate;
    }

    @Override
    public ListIterator<ElementType> iterator() {
        if (delegate != null) {
            // The iterator is a read-only iterator, so we can return it
            // unmodified.
            return delegate.iterator();
        }
        return new ListIterator<ElementType>() {

            private int index = 0;

            @Override
            public boolean hasNext() {
                return (nextIndex() < getValueSize());
            }

            @Override
            public ElementType next() {
                if (!hasNext()) {
                    throw new NoSuchElementException(
                            "This iterator is already at the end of the list.");
                }
                index += 1;
                return getGenericValueElement(index);
            }

            @Override
            public boolean hasPrevious() {
                int previousIndex = previousIndex();
                if (previousIndex < 0) {
                    return false;
                }
                return (previousIndex < getValueSize());
            }

            @Override
            public ElementType previous() {
                if (!hasPrevious()) {
                    throw new NoSuchElementException(
                            "This iterator is already at the beginning of the list.");
                }
                index -= 1;
                return getGenericValueElement(index);
            }

            @Override
            public int nextIndex() {
                return index + 1;
            }

            @Override
            public int previousIndex() {
                return index - 1;
            }

            @Override
            public void remove() {
                throw new UnsupportedOperationException(
                        "This iterator allows read access only.");
            }

            @Override
            public void set(ElementType e) {
                throw new UnsupportedOperationException(
                        "This iterator allows read access only.");
            }

            @Override
            public void add(ElementType e) {
                throw new UnsupportedOperationException(
                        "This iterator allows read access only.");
            }
        };
    }

    @Override
    public boolean isReadOnly() {
        return (delegate != null);
    }

    /**
     * Extracts the delegate from a {@link ChannelAccessValueBase}. If
     * <code>obj</code> is an instance of <code>ChannelAccessValueBase</code>
     * and its {@link #delegate} is not <code>null</code>, the
     * <code>delegate</code> is returned. Otherwise, <code>obj</code> is
     * returned.
     * 
     * @param obj
     *            object to check for a delegate (may be <code>null</code>).
     * @return the value's delegate or <code>obj</code> if <code>obj</code> is
     *         <code>null</code>, <code>obj</code> is not an instance of
     *         <code>ChannelAccessValueBase</code>, and if <code>obj</code>'s
     *         delegate is <code>null</code>.
     */
    protected static Object extractDelegate(Object obj) {
        if (obj != null && obj instanceof ChannelAccessValueBase) {
            ChannelAccessValueBase<?, ?> other = (ChannelAccessValueBase<?, ?>) obj;
            if (other.delegate != null) {
                return other.delegate;
            }
        }
        return obj;
    }

    @Override
    public boolean equals(Object obj) {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        // Child classes will have to override this implementation, because we
        // cannot compare the value array (without knowing the type) and child
        // implementations might have additional fields. However, we can do some
        // basic checks, so that child implementations do not have to implement
        // them.
        if (obj == null) {
            return false;
        }
        // We can compare the types. They have to match exactly because for each
        // specific type, there is only exactly one implementation.
        if (!this.getClass().equals(obj.getClass())) {
            return false;
        }
        return true;
    };

    @Override
    public int hashCode() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        // We cannot generate a hash code that includes the value because we
        // would need to know the array type for this. However, we can generate
        // a hash-code based on the type. This should ensure that derived
        // classes produce different hash codes, even if their representation is
        // the same.
        return new HashCodeBuilder(91, 13).append(getType()).toHashCode();
    }

    private static void addAlarmString(StringBuilder sb,
            ChannelAccessAlarmValue<?> value) {
        sb.append("status=");
        sb.append(value.getAlarmStatus().toString());
        sb.append(", severity=");
        sb.append(value.getAlarmSeverity().toString());
        sb.append(", ");
    }

    private static void addTimeString(StringBuilder sb,
            ChannelAccessTimeValue<?> value) {
        sb.append("time=");
        sb.append(value.getTimeStamp().toString());
        sb.append(", ");
    }

    private static void addGraphicsString(StringBuilder sb,
            ChannelAccessGraphicsValue<?> graphicsValue) {
        if (graphicsValue instanceof ChannelAccessFloatingPointGraphicsValue) {
            ChannelAccessFloatingPointGraphicsValue<?> value = (ChannelAccessFloatingPointGraphicsValue<?>) graphicsValue;
            sb.append("precision=");
            short precision = value.getPrecision();
            sb.append(precision & 0xffff);
            sb.append(", ");
        }
        if (graphicsValue instanceof ChannelAccessNumericGraphicsValue) {
            ChannelAccessNumericGraphicsValue<?> value = (ChannelAccessNumericGraphicsValue<?>) graphicsValue;
            sb.append("units=\"");
            sb.append(StringEscapeUtils.escapeJava(value.getUnits()));
            sb.append("\", upperDisplayLimit=");
            sb.append(value.getGenericUpperDisplayLimit().toString());
            sb.append(", lowerDisplayLimit=");
            sb.append(value.getGenericLowerDisplayLimit().toString());
            sb.append(", upperAlarmLimit=");
            sb.append(value.getGenericUpperAlarmLimit().toString());
            sb.append(", upperWarningLimit=");
            sb.append(value.getGenericUpperWarningLimit().toString());
            sb.append(", lowerWarningLimit=");
            sb.append(value.getGenericLowerWarningLimit().toString());
            sb.append(", lowerAlarmLimit=");
            sb.append(value.getGenericLowerAlarmLimit().toString());
            sb.append(", ");
        }
        if (graphicsValue instanceof ChannelAccessGraphicsEnum) {
            ChannelAccessGraphicsEnum value = (ChannelAccessGraphicsEnum) graphicsValue;
            sb.append("labels=[");
            boolean first = true;
            for (String label : value.getLabels()) {
                if (first) {
                    first = false;
                } else {
                    sb.append(", ");
                }
                sb.append('"');
                sb.append(StringEscapeUtils.escapeJava(label));
                sb.append('"');
            }
            sb.append("], ");
        }
    }

    private static void addControlsString(StringBuilder sb,
            ChannelAccessControlsValue<?> controlsValue) {
        if (controlsValue instanceof ChannelAccessNumericControlsValue<?>) {
            ChannelAccessNumericControlsValue<?> value = (ChannelAccessNumericControlsValue<?>) controlsValue;
            sb.append("upperControlLimit=");
            sb.append(value.getGenericUpperControlLimit().toString());
            sb.append(", lowerControlLimit=");
            sb.append(value.getGenericLowerControlLimit().toString());
            sb.append(", ");
        }
    }

    @Override
    public String toString() {
        if (delegate != null) {
            return delegate.toString();
        }
        StringBuilder sb = new StringBuilder();
        sb.append(getType().toString());
        sb.append('{');
        if (this instanceof ChannelAccessConfigureAcknowledgeTransientAlarms) {
            ChannelAccessConfigureAcknowledgeTransientAlarms value = (ChannelAccessConfigureAcknowledgeTransientAlarms) this;
            sb.append("explicitlyAcknowledgeTransientAlarms=");
            sb.append(value.isExplicitlyAcknowledgeTransientAlarms());
            sb.append(", ");
        } else if (this instanceof ChannelAccessAcknowledgeAlarm) {
            ChannelAccessAcknowledgeAlarm value = (ChannelAccessAcknowledgeAlarm) this;
            sb.append("acknowledgeAlarmSeverity=");
            sb.append(value.getAcknowledgeAlarmSeverity().toString());
            sb.append(", ");
        } else if (this instanceof ChannelAccessAlarmAcknowledgementStatus) {
            ChannelAccessAlarmAcknowledgementStatus value = (ChannelAccessAlarmAcknowledgementStatus) this;
            sb.append("status=");
            sb.append(value.getAlarmStatus().toString());
            sb.append(", severity=");
            sb.append(value.getAlarmSeverity().toString());
            sb.append(", explicitlyAcknowledgeTransientAlarms");
            sb.append(value.isExplicitlyAcknowledgeTransientAlarms());
            sb.append(", unacknowledgedSeverity=");
            sb.append(value.getUnacknowledgedAlarmSeverity().toString());
            sb.append(", ");
        } else {
            if (this instanceof ChannelAccessAlarmValue) {
                addAlarmString(sb, (ChannelAccessAlarmValue<?>) this);
            }
            if (this instanceof ChannelAccessTimeValue) {
                addTimeString(sb, (ChannelAccessTimeValue<?>) this);
            }
            if (this instanceof ChannelAccessGraphicsValue) {
                addGraphicsString(sb, (ChannelAccessGraphicsValue<?>) this);
            }
            if (this instanceof ChannelAccessControlsValue) {
                addControlsString(sb, (ChannelAccessControlsValue<?>) this);
            }
        }
        sb.append("value=");
        int numberOfElements = getValueSize();
        if (numberOfElements == 1) {
            Object element = getGenericValueElement(0);
            if (element instanceof String) {
                String stringElement = (String) element;
                sb.append('"');
                sb.append(StringEscapeUtils.escapeJava(stringElement));
                sb.append('"');
            } else {
                sb.append(element.toString());
            }
        } else {
            sb.append('[');
            boolean first = true;
            for (int i = 0; i < numberOfElements; i++) {
                if (first) {
                    first = false;
                } else {
                    sb.append(", ");
                }
                Object element = getGenericValueElement(i);
                if (element instanceof String) {
                    String stringElement = (String) element;
                    sb.append('"');
                    sb.append(StringEscapeUtils.escapeJava(stringElement));
                    sb.append('"');
                } else {
                    sb.append(element.toString());
                }
            }
            sb.append(']');
        }
        sb.append('}');
        return sb.toString();
    }

    @Override
    public ChannelAccessValue<ElementType> clone() {
        // We do not handle the read-only variant here. If this value is
        // read-only, this method should never be called by the child class.
        assert (!isReadOnly());
        try {
            @SuppressWarnings("unchecked")
            ChannelAccessValue<ElementType> copy = (ChannelAccessValue<ElementType>) super
                    .clone();
            return copy;
        } catch (CloneNotSupportedException e) {
            // We know that we implement clonable, thus this should never
            // happen.
            throw new RuntimeException("Unexpected error: " + e.getMessage(), e);
        }
    }

    /**
     * Serializes this value into its binary representation.
     * 
     * @param byteSink
     *            byte sink to write data to.
     * @param count
     *            number of elements that should be serialized. Must not be
     *            negative and must not be greater than the number of elements
     *            in this value (as returned by {@link #getValueSize()}).
     *            However, these limits are not checked by this method.
     * @throws BufferUnderflowException
     *             if the byte sink throws such an exception.
     */
    public abstract void serialize(ByteSink byteSink, int count);

}
